import warnings
from typing import Type

import pydantic

from datahub.configuration.common import ConfigurationWarning


def pydantic_removed_field(
    field: str,
    print_warning: bool = True,
) -> classmethod:
    def _validate_field_removal(cls: Type, values: dict) -> dict:
        if field in values:
            if print_warning:
                warnings.warn(
                    f"The {field} was removed, please remove it from your recipe.",
                    ConfigurationWarning,
                    stacklevel=2,
                )
            values.pop(field)
        return values

    # Hack: Pydantic maintains unique list of validators by referring its __name__.
    # https://github.com/pydantic/pydantic/blob/v1.10.9/pydantic/main.py#L264
    # This hack ensures that multiple field removals do not overwrite each other.
    _validate_field_removal.__name__ = f"{_validate_field_removal.__name__}_{field}"
    return pydantic.root_validator(pre=True, allow_reuse=True)(_validate_field_removal)



from abc import ABCMeta, abstractmethod
from typing import List


class SqlParserException(Exception):
    """Raised when sql parser fails"""

    pass


class SQLParser(metaclass=ABCMeta):
    def __init__(self, sql_query: str, use_external_process: bool = True) -> None:
        self._sql_query = sql_query

    @abstractmethod
    def get_tables(self) -> List[str]:
        pass

    @abstractmethod
    def get_columns(self) -> List[str]:
        pass



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Callable, Iterable, List, Optional, Union

import pydantic
from pydantic import BaseModel

import datahub.emitter.mce_builder as builder
from datahub.api.entities.corpuser.corpuser import CorpUser, CorpUserGenerationConfig
from datahub.configuration.common import ConfigurationError
from datahub.configuration.validate_field_rename import pydantic_renamed_field
from datahub.emitter.generic_emitter import Emitter
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.rest_emitter import DatahubRestEmitter
from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata.schema_classes import (
    CorpGroupEditableInfoClass,
    CorpGroupInfoClass,
    GroupMembershipClass,
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
    StatusClass,
    _Aspect,
)

logger = logging.getLogger(__name__)


@dataclass
class CorpGroupGenerationConfig:
    """
    A container for configuration for generation of mcp-s from CorpGroup instances
    """

    override_editable: bool = False
    datahub_graph: Optional[DataHubGraph] = None


class CorpGroup(BaseModel):
    """This is a CorpGroup class which represents a CorpGroup

    Args:
        id (str): The id of the group
        display_name (Optional[str]): The name of the group
        email (Optional[str]): email of this group
        description (Optional[str]): A description of the group
        overrideEditable (bool): If True, group information that is editable in the UI will be overridden
        picture_link (Optional[str]): A URL which points to a picture which user wants to set as the photo for the group
        slack (Optional[str]): Slack channel for the group
        owners (List[Union[str, CorpUser]]): A list of owner/administrator ids (or urns) for the group. You can also provide the user record for the owner inline within this section
        members (List[Union[str, CorpUser]]): A list of member ids (or urns) for the group.
    """

    id: str

    # These are for CorpGroupInfo
    display_name: Optional[str] = None
    email: Optional[str] = None
    description: Optional[str] = None

    # These are for CorpGroupEditableInfo
    overrideEditable: bool = False
    picture_link: Optional[str] = None
    slack: Optional[str] = None
    owners: List[Union[str, CorpUser]] = []
    members: List[Union[str, CorpUser]] = []

    _rename_admins_to_owners = pydantic_renamed_field("admins", "owners")

    @pydantic.validator("owners", "members", each_item=True)
    def make_urn_if_needed(v):
        if isinstance(v, str):
            return builder.make_user_urn(v)
        return v

    @property
    def urn(self):
        return builder.make_group_urn(self.id)

    def _needs_editable_aspect(self) -> bool:
        return bool(self.picture_link)

    def generate_mcp(
        self, generation_config: CorpGroupGenerationConfig = CorpGroupGenerationConfig()
    ) -> Iterable[MetadataChangeProposalWrapper]:
        urns_created = set()  # dedup member creation on the way out
        members_to_create: List[CorpUser] = (
            [u for u in self.members if isinstance(u, CorpUser)] if self.members else []
        )
        owners_to_create: List[CorpUser] = (
            [u for u in self.owners if isinstance(u, CorpUser)] if self.owners else []
        )

        member_urns: List[str] = (
            [u.urn if isinstance(u, CorpUser) else u for u in self.members]
            if self.members
            else []
        )
        owner_urns: List[str] = (
            [u.urn if isinstance(u, CorpUser) else u for u in self.owners]
            if self.owners
            else []
        )

        for m in members_to_create + owners_to_create:
            if m.urn not in urns_created:
                yield from m.generate_mcp(
                    generation_config=CorpUserGenerationConfig(
                        override_editable=generation_config.override_editable
                    )
                )
                urns_created.add(m.urn)
            else:
                logger.warn(
                    f"Suppressing emission of member {m.urn} before we already emitted metadata for it"
                )

        aspects: List[_Aspect] = [StatusClass(removed=False)]
        if generation_config.override_editable:
            aspects.append(
                CorpGroupEditableInfoClass(
                    description=self.description,
                    pictureLink=self.picture_link,
                    slack=self.slack,
                    email=self.email,
                )
            )
        else:
            aspects.append(
                CorpGroupInfoClass(
                    admins=owner_urns,  # deprecated but we fill it out for consistency
                    members=member_urns,  # deprecated but we fill it out for consistency
                    groups=[],  # deprecated
                    displayName=self.display_name,
                    email=self.email,
                    description=self.description,
                    slack=self.slack,
                )
            )
            # picture link is only available in the editable aspect, so we have to use it if it is provided
            if self._needs_editable_aspect():
                aspects.append(
                    CorpGroupEditableInfoClass(
                        description=self.description,
                        pictureLink=self.picture_link,
                        slack=self.slack,
                        email=self.email,
                    )
                )
        for aspect in aspects:
            yield MetadataChangeProposalWrapper(entityUrn=self.urn, aspect=aspect)

        # Add owners to the group.
        if owner_urns:
            ownership = OwnershipClass(owners=[])
            for urn in owner_urns:
                ownership.owners.append(
                    OwnerClass(owner=urn, type=OwnershipTypeClass.TECHNICAL_OWNER)
                )
            yield MetadataChangeProposalWrapper(entityUrn=self.urn, aspect=ownership)

        # Unfortunately, the members in CorpGroupInfo has been deprecated.
        # So we need to emit GroupMembership oriented to the individual users.
        # TODO: Move this to PATCH MCP-s once these aspects are supported via patch.
        if generation_config.datahub_graph is not None:
            datahub_graph = generation_config.datahub_graph

            # Add group membership to each user.
            for urn in member_urns:
                group_membership = datahub_graph.get_aspect(
                    urn, GroupMembershipClass
                ) or GroupMembershipClass(groups=[])
                if self.urn not in group_membership.groups:
                    group_membership.groups = sorted(
                        set(group_membership.groups + [self.urn])
                    )
                    yield MetadataChangeProposalWrapper(
                        entityUrn=urn, aspect=group_membership
                    )
        else:
            if member_urns:
                raise ConfigurationError(
                    "Unable to emit group membership because members is non-empty, and a DataHubGraph instance was not provided."
                )

        # emit status aspects for all user urns referenced (to ensure they get created)
        for urn in set(owner_urns).union(set(member_urns)):
            yield MetadataChangeProposalWrapper(
                entityUrn=urn, aspect=StatusClass(removed=False)
            )

    def emit(
        self,
        emitter: Emitter,
        callback: Optional[Callable[[Exception, str], None]] = None,
    ) -> None:
        """
        A convenience method to emit the CorpGroup entity to DataHub using an emitter.
        See also: generate_mcp to have finer grain control over mcp routing

        :param emitter: Datahub Emitter to emit the proccess event
        :param callback: The callback method for KafkaEmitter if it is used
        """
        datahub_graph = emitter if isinstance(emitter, DataHubGraph) else None
        if not datahub_graph:
            if isinstance(emitter, DatahubRestEmitter):
                # create a datahub graph instance from the emitter
                # this code path exists mainly for backwards compatibility with existing callers
                # who are passing in a DataHubRestEmitter today
                # we won't need this in the future once PATCH support is implemented as all emitters
                # will work
                datahub_graph = emitter.to_graph()
        for mcp in self.generate_mcp(
            generation_config=CorpGroupGenerationConfig(
                override_editable=self.overrideEditable, datahub_graph=datahub_graph
            )
        ):
            emitter.emit(mcp, callback)



import re
from enum import Enum
from typing import Generator, List, Tuple

CONTROL_FLOW_KEYWORDS = [
    "GO",
    r"BEGIN\w+TRY",
    r"BEGIN\w+CATCH",
    "BEGIN",
    r"END\w+TRY",
    r"END\w+CATCH",
    "END",
]

# There's an exception to this rule, which is when the statement
# is preceeded by a CTE.
FORCE_NEW_STATEMENT_KEYWORDS = [
    # SELECT is used inside queries as well, so we can't include it here.
    "INSERT",
    "UPDATE",
    "DELETE",
    "MERGE",
]


class ParserState(Enum):
    NORMAL = 1
    STRING = 2
    COMMENT = 3
    MULTILINE_COMMENT = 4


def _is_keyword_at_position(sql: str, pos: int, keyword: str) -> bool:
    """
    Check if a keyword exists at the given position using regex word boundaries.
    """
    if pos + len(keyword) > len(sql):
        return False

    # If we're not at a word boundary, we can't generate a keyword.
    if pos > 0 and not (
        bool(re.match(r"\w\W", sql[pos - 1 : pos + 1]))
        or bool(re.match(r"\W\w", sql[pos - 1 : pos + 1]))
    ):
        return False

    pattern = rf"^{re.escape(keyword)}\b"
    match = re.match(pattern, sql[pos:], re.IGNORECASE)
    return bool(match)


def _look_ahead_for_keywords(
    sql: str, pos: int, keywords: List[str]
) -> Tuple[bool, str, int]:
    """
    Look ahead for SQL keywords at the current position.
    """

    for keyword in keywords:
        if _is_keyword_at_position(sql, pos, keyword):
            return True, keyword, len(keyword)
    return False, "", 0


def split_statements(sql: str) -> Generator[str, None, None]:
    """
    Split T-SQL code into individual statements, handling various SQL constructs.
    """
    if not sql or not sql.strip():
        return

    current_statement: List[str] = []
    state = ParserState.NORMAL
    i = 0

    def yield_if_complete() -> Generator[str, None, None]:
        statement = "".join(current_statement).strip()
        if statement:
            yield statement
            current_statement.clear()

    prev_real_char = "\0"  # the most recent non-whitespace, non-comment character
    while i < len(sql):
        c = sql[i]
        next_char = sql[i + 1] if i < len(sql) - 1 else "\0"

        if state == ParserState.NORMAL:
            if c == "'":
                state = ParserState.STRING
                current_statement.append(c)
                prev_real_char = c
            elif c == "-" and next_char == "-":
                state = ParserState.COMMENT
                current_statement.append(c)
                current_statement.append(next_char)
                i += 1
            elif c == "/" and next_char == "*":
                state = ParserState.MULTILINE_COMMENT
                current_statement.append(c)
                current_statement.append(next_char)
                i += 1
            else:
                most_recent_real_char = prev_real_char
                if not c.isspace():
                    prev_real_char = c

                is_control_keyword, keyword, keyword_len = _look_ahead_for_keywords(
                    sql, i, keywords=CONTROL_FLOW_KEYWORDS
                )
                if is_control_keyword:
                    # Yield current statement if any
                    yield from yield_if_complete()
                    # Yield keyword as its own statement
                    yield keyword
                    i += keyword_len
                    continue

                (
                    is_force_new_statement_keyword,
                    keyword,
                    keyword_len,
                ) = _look_ahead_for_keywords(
                    sql, i, keywords=FORCE_NEW_STATEMENT_KEYWORDS
                )
                if (
                    is_force_new_statement_keyword and most_recent_real_char != ")"
                ):  # usually we'd have a close paren that closes a CTE
                    # Force termination of current statement
                    yield from yield_if_complete()

                    current_statement.append(keyword)
                    i += keyword_len
                    continue

                elif c == ";":
                    yield from yield_if_complete()
                else:
                    current_statement.append(c)

        elif state == ParserState.STRING:
            current_statement.append(c)
            if c == "'" and next_char == "'":
                current_statement.append(next_char)
                i += 1
            elif c == "'":
                state = ParserState.NORMAL

        elif state == ParserState.COMMENT:
            current_statement.append(c)
            if c == "\n":
                state = ParserState.NORMAL

        elif state == ParserState.MULTILINE_COMMENT:
            current_statement.append(c)
            if c == "*" and next_char == "/":
                current_statement.append(next_char)
                i += 1
                state = ParserState.NORMAL

        i += 1

    # Handle the last statement
    yield from yield_if_complete()



import functools
import json
import logging
import os
import tempfile
import unittest
import urllib.request
from dataclasses import dataclass
from os.path import basename, dirname
from pathlib import Path
from typing import Any, Iterable, List, Optional, Union
from urllib.parse import urlparse

import jsonref
from pydantic import AnyHttpUrl, DirectoryPath, FilePath, validator
from pydantic.fields import Field

import datahub.metadata.schema_classes as models
from datahub.configuration.common import ConfigModel
from datahub.configuration.source_common import DatasetSourceConfigMixin
from datahub.emitter.mce_builder import (
    make_dataplatform_instance_urn,
    make_dataset_urn_with_platform_instance,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (  # SourceCapability,; capability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor, SourceCapability
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.extractor.json_ref_patch import title_swapping_callback
from datahub.ingestion.extractor.json_schema_util import (
    JsonSchemaTranslator,
    get_schema_metadata,
)
from datahub.ingestion.source.common.subtypes import DatasetSubTypes
from datahub.ingestion.source.state.entity_removal_state import GenericCheckpointState
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
    StaleEntityRemovalSourceReport,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
    StatefulIngestionSourceBase,
)
from datahub.utilities.urns.data_platform_urn import DataPlatformUrn

logger = logging.getLogger(__name__)


def is_url_valid(url: Optional[str]) -> bool:
    if url is None:
        return False
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except Exception:
        return False


class URIReplacePattern(ConfigModel):
    match: str = Field(
        description="Pattern to match on uri-s as part of reference resolution. See replace field",
    )
    replace: str = Field(
        description="Pattern to replace with as part of reference resolution. See match field",
    )


class JsonSchemaSourceConfig(StatefulIngestionConfigBase, DatasetSourceConfigMixin):
    path: Union[FilePath, DirectoryPath, AnyHttpUrl] = Field(
        description="Set this to a single file-path or a directory-path (for recursive traversal) or a remote url. e.g. https://json.schemastore.org/petstore-v1.0.json"
    )
    platform: str = Field(
        description="Set this to a platform that you want all schemas to live under. e.g. schemaregistry / schemarepo etc."
    )
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = None
    use_id_as_base_uri: bool = Field(
        default=False,
        description="When enabled, uses the `$id` field in the json schema as the base uri for following references.",
    )
    uri_replace_pattern: Optional[URIReplacePattern] = Field(
        default=None,
        description="Use this if URI-s need to be modified during reference resolution. Simple string match - replace capabilities are supported.",
    )

    @validator("path")
    def download_http_url_to_temp_file(v):
        if isinstance(v, AnyHttpUrl):
            try:
                with urllib.request.urlopen(v) as response:
                    schema_dict = json.load(response)
                    if not JsonSchemaTranslator._get_id_from_any_schema(schema_dict):
                        schema_dict["$id"] = str(v)
                    with tempfile.NamedTemporaryFile(
                        mode="w", delete=False
                    ) as tmp_file:
                        tmp_file.write(json.dumps(schema_dict))
                        tmp_file.flush()
                        return tmp_file.name
            except Exception as e:
                logger.error(
                    f"Failed to localize url {v} due to {e}. Run with --debug to get full stacktrace"
                )
                logger.debug(f"Failed to localize url {v} due to {e}", exc_info=e)
                raise
        return v


class JsonSchemaCheckpointState(GenericCheckpointState):
    def __init__(self, **data: Any):
        super().__init__(**data)


@platform_name(platform_name="JSON Schemas", id="json-schema")
@config_class(JsonSchemaSourceConfig)
@support_status(SupportStatus.INCUBATING)
@capability(
    capability_name=SourceCapability.SCHEMA_METADATA,
    description="Extracts schemas, following references",
)
@capability(
    capability_name=SourceCapability.PLATFORM_INSTANCE,
    description="Supports platform instance via config",
)
@capability(
    capability_name=SourceCapability.DESCRIPTIONS,
    description="Extracts descriptions at top level and field level",
)
@capability(
    capability_name=SourceCapability.DELETION_DETECTION,
    description="With stateful ingestion enabled, will remove entities from DataHub if they are no longer present in the source",
)
@capability(
    capability_name=SourceCapability.TAGS,
    description="Does not currently support extracting tags",
    supported=False,
)
@capability(
    capability_name=SourceCapability.OWNERSHIP,
    description="Does not currently support extracting ownership",
    supported=False,
)
@dataclass
class JsonSchemaSource(StatefulIngestionSourceBase):
    """
    This source extracts metadata from a single JSON Schema or multiple JSON Schemas rooted at a particular path.
    It performs reference resolution based on the `$ref` keyword.

    Metadata mapping:
    - Schemas are mapped to Datasets with sub-type Schema
    - The name of the Schema (Dataset) is inferred from the `$id` property and if that is missing, the file name.
    - Browse paths are minted based on the path
    """

    @staticmethod
    def _pull_up_reference(schema_dict):
        """Transforms a schema that is of the form
        {
         "$ref": "#definitions/MyDefinition",
         "definitions": {
            "MyDefinition": {
                ... definition
            },
            "OtherDefinition": {
                ... other definition
            }
         }
        }
        to
        {
            "title": "MyDefinition",
            ... definition

            "definitions": {
            "OtherDefinition": {
                ... other definition
            }
         }
        """

        if "$ref" in schema_dict:
            reference = schema_dict["$ref"]
            if reference.startswith("#/"):
                parts = reference[2:].split("/")
                referred = schema_dict
                for p in parts:
                    referred = referred[p]
                # copy referred object into the top level
                for k, v in referred.items():
                    schema_dict[k] = v
                # remove object from definitions dict
                deletable = schema_dict
                for p in parts[:-1]:
                    deletable = deletable[p]
                deletable.pop(parts[-1])
                # remove the "$ref" key
                schema_dict.pop("$ref")
                # set the title to reflect the definition name
                if "title" not in schema_dict:
                    schema_dict["title"] = parts[-1]
        return schema_dict

    @staticmethod
    def _load_json_schema(filename, loader, use_id_as_base_uri):
        """Loads the given schema file"""
        path = Path(filename).resolve()
        base_path = dirname(str(path))
        base_uri = f"file://{base_path}/"

        with open(path) as schema_file:
            logger.info(f"Opening file {path}")
            schema_dict = json.load(schema_file)
            try:
                schema_dict = JsonSchemaSource._pull_up_reference(schema_dict)
                jsonref_dict = jsonref.loads(
                    json.dumps(schema_dict),
                    base_uri=base_uri,
                    jsonschema=use_id_as_base_uri,
                    loader=loader,
                    lazy_load=False,
                )
                resolved_dict = jsonref_dict
            except Exception:
                resolved_dict = schema_dict
            return (resolved_dict, json.dumps(schema_dict))

    @staticmethod
    def stringreplaceloader(match_string, replace_string, uri, **kwargs):
        """
        Provides a callable which takes a URI, and returns the loaded JSON referred
        to by that URI. Uses :mod:`requests` if available for HTTP URIs, and falls
        back to :mod:`urllib`.
        """
        uri = uri.replace(match_string, replace_string)
        logger.debug(f"Resolving ref to file {uri}")
        return jsonref.jsonloader(uri, **kwargs)

    def __init__(self, ctx: PipelineContext, config: JsonSchemaSourceConfig):
        super().__init__(ctx=ctx, config=config)
        self.config = config
        self.report = StaleEntityRemovalSourceReport()

    def _load_one_file(
        self, ref_loader: Any, browse_prefix: str, root_dir: Path, file_name: str
    ) -> Iterable[MetadataWorkUnit]:
        with unittest.mock.patch("jsonref.JsonRef.callback", title_swapping_callback):
            (schema_dict, schema_string) = self._load_json_schema(
                os.path.join(root_dir, file_name),
                loader=ref_loader,
                use_id_as_base_uri=self.config.use_id_as_base_uri,
            )
            logger.debug(f"Loaded file {file_name} as {schema_dict}")
            if not JsonSchemaTranslator._get_id_from_any_schema(schema_dict):
                schema_type = os.path.basename(file_name)[: -len(".json")]
            else:
                schema_id = JsonSchemaTranslator._get_id_from_any_schema(schema_dict)
                assert schema_id
                schema_type = schema_id.replace("https:/", "")

            if schema_type.endswith(".json"):
                schema_type = schema_type[: -len(".json")]

            browse_path = browse_prefix + dirname(schema_type)
            dataset_simple_name = basename(schema_type)

            dataset_name = schema_type

            meta: models.SchemaMetadataClass = get_schema_metadata(
                platform=self.config.platform,
                name=dataset_name,
                json_schema=schema_dict,
                raw_schema_string=schema_string,
            )
            dataset_urn = make_dataset_urn_with_platform_instance(
                platform=self.config.platform,
                name=dataset_name,
                platform_instance=self.config.platform_instance,
                env=self.config.env,
            )
            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn, aspect=meta
            ).as_workunit()

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn, aspect=models.StatusClass(removed=False)
            ).as_workunit()

            external_url = JsonSchemaTranslator._get_id_from_any_schema(schema_dict)
            if not is_url_valid(external_url):
                external_url = None

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=models.DatasetPropertiesClass(
                    externalUrl=external_url,
                    name=dataset_simple_name,
                    description=JsonSchemaTranslator._get_description_from_any_schema(
                        schema_dict
                    ),
                ),
            ).as_workunit()

            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=models.SubTypesClass(typeNames=[DatasetSubTypes.SCHEMA]),
            ).as_workunit()

            browse_path = browse_prefix + dirname(schema_type)
            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=models.BrowsePathsClass(paths=[browse_path]),
            ).as_workunit()

            if self.config.platform_instance:
                yield MetadataChangeProposalWrapper(
                    entityUrn=dataset_urn,
                    aspect=models.DataPlatformInstanceClass(
                        platform=str(
                            DataPlatformUrn.create_from_id(self.config.platform)
                        ),
                        instance=make_dataplatform_instance_urn(
                            platform=self.config.platform,
                            instance=self.config.platform_instance,
                        ),
                    ),
                ).as_workunit()

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        if self.config.uri_replace_pattern:
            ref_loader = functools.partial(
                self.stringreplaceloader,
                self.config.uri_replace_pattern.match,
                self.config.uri_replace_pattern.replace,
            )
        else:
            ref_loader = jsonref.jsonloader
        browse_prefix = f"/{self.config.env.lower()}/{self.config.platform}"
        if self.config.platform_instance:
            browse_prefix = f"/{self.config.env.lower()}/{self.config.platform}/{self.config.platform_instance}"

        if os.path.isdir(self.config.path):
            for root, dirs, files in os.walk(self.config.path, topdown=False):
                for file_name in [f for f in files if f.endswith(".json")]:
                    try:
                        yield from self._load_one_file(
                            ref_loader,
                            browse_prefix=browse_prefix,
                            root_dir=Path(root),
                            file_name=file_name,
                        )
                    except Exception as e:
                        self.report.report_failure(
                            f"{root}/{file_name}", f"Failed to process due to {e}"
                        )
                        logger.error(
                            f"Failed to process file {root}/{file_name}", exc_info=e
                        )

        else:
            try:
                yield from self._load_one_file(
                    ref_loader,
                    browse_prefix=browse_prefix,
                    root_dir=Path(os.path.dirname(Path(self.config.path))),
                    file_name=str(self.config.path),
                )
            except Exception as e:
                self.report.report_failure(
                    str(self.config.path), f"Failed to process due to {e}"
                )
                logger.error(f"Failed to process file {self.config.path}", exc_info=e)

    def get_report(self):
        return self.report



import logging
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, validator

from datahub.ingestion.graph.client import DatahubClientConfig, DataHubGraph
from datahub.secret.datahub_secrets_client import DataHubSecretsClient
from datahub.secret.secret_store import SecretStore

logger = logging.getLogger(__name__)


class DataHubSecretStoreConfig(BaseModel):
    graph_client: Optional[DataHubGraph] = None
    graph_client_config: Optional[DatahubClientConfig] = None

    class Config:
        arbitrary_types_allowed = True

    @validator("graph_client")
    def check_graph_connection(cls, v: DataHubGraph) -> DataHubGraph:
        if v is not None:
            v.test_connection()
        return v


# An implementation of SecretStore that fetches secrets from DataHub
class DataHubSecretStore(SecretStore):
    # Client for fetching secrets from DataHub GraphQL API
    client: DataHubSecretsClient

    def __init__(self, config: DataHubSecretStoreConfig):
        # Attempt to establish an outbound connection to DataHub and create a client.
        if config.graph_client is not None:
            self.client = DataHubSecretsClient(graph=config.graph_client)
        elif config.graph_client_config is not None:
            graph = DataHubGraph(config.graph_client_config)
            self.client = DataHubSecretsClient(graph)
        else:
            raise Exception(
                "Invalid configuration provided: unable to construct DataHub Graph Client."
            )

    def get_secret_values(self, secret_names: List[str]) -> Dict[str, Union[str, None]]:
        # Fetch the secret from DataHub, using the credentials provided in the configuration.
        # Use the GraphQL API.
        try:
            return self.client.get_secret_values(secret_names)
        except Exception:
            # Failed to resolve secrets, return empty.
            logger.exception(
                f"Caught exception while attempting to fetch secrets from DataHub. Secret names: {secret_names}"
            )
            return {}

    def get_secret_value(self, secret_name: str) -> Union[str, None]:
        secret_value_dict = self.get_secret_values([secret_name])
        return secret_value_dict.get(secret_name)

    def get_id(self) -> str:
        return "datahub"

    @classmethod
    def create(cls, config: Any) -> "DataHubSecretStore":
        config = DataHubSecretStoreConfig.parse_obj(config)
        return cls(config)



from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Dict, Generic, Iterable, List, Optional, Type, TypeVar, Union, cast

from datahub.configuration.common import ConfigModel
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.closeable import Closeable
from datahub.metadata.schema_classes import MetadataChangeEventClass, _Aspect
from datahub.utilities.type_annotations import get_class_from_annotation

LiteConfig = TypeVar("LiteConfig", bound=ConfigModel)


class AutoComplete(ConfigModel):
    success_path: str
    failed_token: str
    suggested_path: str


class Browseable(ConfigModel):
    id: str
    name: str
    leaf: bool = False
    parents: Optional[List[str]] = None
    auto_complete: Optional[AutoComplete] = None


class Searchable(ConfigModel):
    id: str
    aspect: Optional[str]
    snippet: Optional[str]


class SearchFlavor(Enum):
    FREE_TEXT = "free-text"
    EXACT = "exact"


class PathNotFoundException(Exception):
    pass


class DataHubLiteLocal(Generic[LiteConfig], Closeable, metaclass=ABCMeta):
    """The embedded version of DataHub Lite. All DataHub Lite implementations need to extend this class"""

    def __init__(self, config: LiteConfig):
        pass

    @classmethod
    def get_config_class(cls) -> Type[LiteConfig]:
        config_class = get_class_from_annotation(cls, DataHubLiteLocal, ConfigModel)
        assert config_class, "DataHubLiteLocal subclasses must define a config class"
        return cast(Type[LiteConfig], config_class)

    @abstractmethod
    def location(self) -> str:
        pass

    @abstractmethod
    def destroy(self) -> None:
        pass

    @abstractmethod
    def write(
        self,
        record: Union[
            MetadataChangeEventClass,
            MetadataChangeProposalWrapper,
        ],
    ) -> None:
        pass

    @abstractmethod
    def list_ids(self) -> Iterable[str]:
        pass

    @abstractmethod
    def get(
        self,
        id: str,
        aspects: Optional[List[str]],
        typed: bool = False,
        as_of: Optional[int] = None,
        details: Optional[bool] = False,
    ) -> Optional[Dict[str, Union[str, dict, _Aspect]]]:
        pass

    @abstractmethod
    def search(
        self,
        query: str,
        flavor: SearchFlavor,
        aspects: List[str] = [],
        snippet: bool = True,
    ) -> Iterable[Searchable]:
        pass

    @abstractmethod
    def ls(self, path: str) -> List[Browseable]:
        pass

    @abstractmethod
    def get_all_entities(
        self, typed: bool = False
    ) -> Iterable[Dict[str, Union[dict, _Aspect]]]:
        pass

    @abstractmethod
    def get_all_aspects(self) -> Iterable[MetadataChangeProposalWrapper]:
        pass

    @abstractmethod
    def reindex(self) -> None:
        pass



from typing import Dict, List, Optional, Set, cast

from datahub.configuration.common import (
    TransformerSemantics,
    TransformerSemanticsConfigModel,
)
from datahub.emitter.mce_builder import Aspect
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.transformer.dataset_domain import AddDatasetDomain
from datahub.ingestion.transformer.dataset_transformer import DatasetDomainTransformer
from datahub.metadata.schema_classes import DomainsClass, GlobalTagsClass


class DatasetTagDomainMapperConfig(TransformerSemanticsConfigModel):
    domain_mapping: Dict[str, str]


class DatasetTagDomainMapper(DatasetDomainTransformer):
    """A transformer that appends a predefined set of domains to each dataset that includes specific tags defined in the transformer."""

    def __init__(self, config: DatasetTagDomainMapperConfig, ctx: PipelineContext):
        super().__init__()
        self.ctx: PipelineContext = ctx
        self.config: DatasetTagDomainMapperConfig = config

    @classmethod
    def create(
        cls, config_dict: dict, ctx: PipelineContext
    ) -> "DatasetTagDomainMapper":
        config = DatasetTagDomainMapperConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[Aspect]
    ) -> Optional[Aspect]:
        # Initialize the existing domain aspect
        existing_domain_aspect: DomainsClass = cast(DomainsClass, aspect)
        assert self.ctx.graph
        global_tags: Optional[GlobalTagsClass] = self.ctx.graph.get_tags(entity_urn)
        # Check if we have tags received in existing aspect
        if global_tags:
            domain_mapping = self.config.domain_mapping
            transformer_tags = domain_mapping.keys()
            tags_seen: Set[str] = set()
            for tag_item in global_tags.tags:
                tag = tag_item.tag.split("urn:li:tag:")[-1]
                if tag in transformer_tags:
                    tags_seen.add(tag)

            if tags_seen:
                domain_aspect = DomainsClass(domains=[])
                domains_to_add: List[str] = []
                for tag in tags_seen:
                    if domain_mapping.get(tag):
                        domains_to_add.append(domain_mapping[tag])

                mapped_domains = AddDatasetDomain.get_domain_class(
                    self.ctx.graph, domains_to_add
                )
                domain_aspect.domains.extend(mapped_domains.domains)
                if self.config.semantics == TransformerSemantics.PATCH:
                    # Try merging with server-side domains
                    patch_domain_aspect: Optional[
                        DomainsClass
                    ] = AddDatasetDomain._merge_with_server_domains(
                        self.ctx.graph, entity_urn, domain_aspect
                    )
                    return cast(Optional[Aspect], patch_domain_aspect)
                return cast(Optional[Aspect], domain_aspect)
        return cast(Optional[Aspect], existing_domain_aspect)



import json
import logging
import time
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generic, Iterable, List, Optional, Set, TypeVar

import pyspark
from databricks.sdk.service.sql import QueryStatementType
from sqllineage.runner import LineageRunner

from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.source_helpers import auto_empty_dataset_usage_statistics
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.unity.config import UnityCatalogSourceConfig
from datahub.ingestion.source.unity.proxy import UnityCatalogApiProxy
from datahub.ingestion.source.unity.proxy_types import (
    OPERATION_STATEMENT_TYPES,
    Query,
    TableReference,
)
from datahub.ingestion.source.unity.report import UnityCatalogReport
from datahub.ingestion.source.usage.usage_common import UsageAggregator
from datahub.metadata.schema_classes import OperationClass
from datahub.sql_parsing.sqlglot_utils import get_query_fingerprint

logger = logging.getLogger(__name__)


TableMap = Dict[str, List[TableReference]]
T = TypeVar("T")


@dataclass  # Dataclass over NamedTuple to support generic type annotations
class GenericTableInfo(Generic[T]):
    source_tables: List[T]
    target_tables: List[T]


StringTableInfo = GenericTableInfo[str]
QueryTableInfo = GenericTableInfo[TableReference]


@dataclass(eq=False)
class UnityCatalogUsageExtractor:
    config: UnityCatalogSourceConfig
    report: UnityCatalogReport
    proxy: UnityCatalogApiProxy
    table_urn_builder: Callable[[TableReference], str]
    user_urn_builder: Callable[[str], str]

    def __post_init__(self):
        self.usage_aggregator = UsageAggregator[TableReference](self.config)
        self._spark_sql_parser: Optional[Any] = None

    @property
    def spark_sql_parser(self):
        """Lazily initializes the Spark SQL parser."""
        if self._spark_sql_parser is None:
            spark_context = pyspark.SparkContext.getOrCreate()
            spark_session = pyspark.sql.SparkSession(spark_context)
            self._spark_sql_parser = (
                spark_session._jsparkSession.sessionState().sqlParser()
            )
        return self._spark_sql_parser

    def get_usage_workunits(
        self, table_refs: Set[TableReference]
    ) -> Iterable[MetadataWorkUnit]:
        try:
            yield from self._get_workunits_internal(table_refs)
        except Exception as e:
            logger.error("Error processing usage", exc_info=True)
            self.report.report_warning("usage-extraction", str(e))

    def _get_workunits_internal(
        self, table_refs: Set[TableReference]
    ) -> Iterable[MetadataWorkUnit]:
        table_map = defaultdict(list)
        query_hashes = set()
        for ref in table_refs:
            table_map[ref.table].append(ref)
            table_map[f"{ref.schema}.{ref.table}"].append(ref)
            table_map[ref.qualified_table_name].append(ref)

        with self.report.usage_perf_report.get_queries_timer as current_timer:
            for query in self._get_queries():
                self.report.num_queries += 1
                with current_timer.pause():
                    with self.report.usage_perf_report.query_fingerprinting_timer:
                        query_hashes.add(
                            get_query_fingerprint(
                                query.query_text, "databricks", fast=True
                            )
                        )
                        self.report.num_unique_queries = len(query_hashes)
                    table_info = self._parse_query(query, table_map)
                    if table_info is not None:
                        if self.config.include_operational_stats:
                            yield from self._generate_operation_workunit(
                                query, table_info
                            )
                        for source_table in table_info.source_tables:
                            with self.report.usage_perf_report.aggregator_add_event_timer:
                                self.usage_aggregator.aggregate_event(
                                    resource=source_table,
                                    start_time=query.start_time,
                                    query=query.query_text,
                                    user=query.user_name,
                                    fields=[],
                                )

        if not self.report.num_queries:
            logger.warning("No queries found in the given time range.")
            self.report.report_warning(
                "usage",
                f"No queries found: "
                f"are you missing the CAN_MANAGE permission on SQL warehouse {self.proxy.warehouse_id}?",
            )
            return

        yield from auto_empty_dataset_usage_statistics(
            self.usage_aggregator.generate_workunits(
                resource_urn_builder=self.table_urn_builder,
                user_urn_builder=self.user_urn_builder,
            ),
            dataset_urns={self.table_urn_builder(ref) for ref in table_refs},
            config=self.config,
        )

    def _generate_operation_workunit(
        self, query: Query, table_info: QueryTableInfo
    ) -> Iterable[MetadataWorkUnit]:
        with self.report.usage_perf_report.gen_operation_timer:
            if (
                not query.statement_type
                or query.statement_type not in OPERATION_STATEMENT_TYPES
            ):
                return None

            # Not sure about behavior when there are multiple target tables. This is a best attempt.
            for target_table in table_info.target_tables:
                operation_aspect = OperationClass(
                    timestampMillis=int(time.time() * 1000),
                    lastUpdatedTimestamp=int(query.end_time.timestamp() * 1000),
                    actor=(
                        self.user_urn_builder(query.user_name)
                        if query.user_name
                        else None
                    ),
                    operationType=OPERATION_STATEMENT_TYPES[query.statement_type],
                    affectedDatasets=[
                        self.table_urn_builder(table)
                        for table in table_info.source_tables
                    ],
                )
                self.report.num_operational_stats_workunits_emitted += 1
                yield MetadataChangeProposalWrapper(
                    entityUrn=self.table_urn_builder(target_table),
                    aspect=operation_aspect,
                ).as_workunit()

    def _get_queries(self) -> Iterable[Query]:
        try:
            yield from self.proxy.query_history(
                self.config.start_time, self.config.end_time
            )
        except Exception as e:
            logger.warning("Error getting queries", exc_info=True)
            self.report.report_warning("get-queries", str(e))

    def _parse_query(
        self, query: Query, table_map: TableMap
    ) -> Optional[QueryTableInfo]:
        with self.report.usage_perf_report.sql_parsing_timer:
            table_info = self._parse_query_via_lineage_runner(query.query_text)
            if table_info is None and query.statement_type == QueryStatementType.SELECT:
                with self.report.usage_perf_report.spark_sql_parsing_timer:
                    table_info = self._parse_query_via_spark_sql_plan(query.query_text)

            if table_info is None:
                self.report.num_queries_dropped_parse_failure += 1
                return None
            else:
                return QueryTableInfo(
                    source_tables=self._resolve_tables(
                        table_info.source_tables, table_map
                    ),
                    target_tables=self._resolve_tables(
                        table_info.target_tables, table_map
                    ),
                )

    def _parse_query_via_lineage_runner(self, query: str) -> Optional[StringTableInfo]:
        try:
            runner = LineageRunner(query)
            return GenericTableInfo(
                source_tables=[
                    self._parse_sqllineage_table(table)
                    for table in runner.source_tables
                ],
                target_tables=[
                    self._parse_sqllineage_table(table)
                    for table in runner.target_tables
                ],
            )
        except Exception as e:
            logger.info(f"Could not parse query via lineage runner, {query}: {e!r}")
            return None

    @staticmethod
    def _parse_sqllineage_table(sqllineage_table: object) -> str:
        full_table_name = str(sqllineage_table)
        default_schema = "<default>."
        if full_table_name.startswith(default_schema):
            return full_table_name[len(default_schema) :]
        else:
            return full_table_name

    def _parse_query_via_spark_sql_plan(self, query: str) -> Optional[StringTableInfo]:
        """Parse query source tables via Spark SQL plan. This is a fallback option."""
        # Would be more effective if we upgrade pyspark
        # Does not work with CTEs or non-SELECT statements
        try:
            plan = json.loads(self.spark_sql_parser.parsePlan(query).toJSON())
            tables = [self._parse_plan_item(item) for item in plan]
            self.report.num_queries_parsed_by_spark_plan += 1
            return GenericTableInfo(
                source_tables=[t for t in tables if t], target_tables=[]
            )
        except Exception as e:
            logger.info(f"Could not parse query via spark plan, {query}: {e!r}")
            return None

    @staticmethod
    def _parse_plan_item(item: dict) -> Optional[str]:
        if item["class"] == "org.apache.spark.sql.catalyst.analysis.UnresolvedRelation":
            return ".".join(item["multipartIdentifier"].strip("[]").split(", "))
        return None

    def _resolve_tables(
        self, tables: List[str], table_map: TableMap
    ) -> List[TableReference]:
        """Resolve tables to TableReferences, filtering out unrecognized or unresolvable table names."""

        missing_table = False
        duplicate_table = False
        output = []
        for table in tables:
            table = str(table)
            if table not in table_map:
                logger.debug(f"Dropping query with unrecognized table: {table}")
                missing_table = True
            else:
                refs = table_map[table]
                if len(refs) == 1:
                    output.append(refs[0])
                else:
                    logger.warning(
                        f"Could not resolve table ref for {table}: {len(refs)} duplicates."
                    )
                    duplicate_table = True

        if missing_table:
            self.report.num_queries_missing_table += 1
        if duplicate_table:
            self.report.num_queries_duplicate_table += 1

        return output



import logging
from dataclasses import dataclass
from datetime import datetime
from random import choices
from typing import Dict, List, Optional

import click
import humanfriendly
import progressbar
from click_default_group import DefaultGroup
from tabulate import tabulate

from datahub.cli import cli_utils
from datahub.configuration.datetimes import ClickDatetime
from datahub.emitter.aspect import ASPECT_MAP, TIMESERIES_ASPECT_MAP
from datahub.ingestion.graph.client import DataHubGraph, get_default_graph
from datahub.ingestion.graph.filters import RemovedStatusFilter
from datahub.telemetry import telemetry
from datahub.upgrade import upgrade
from datahub.utilities.perf_timer import PerfTimer
from datahub.utilities.urns.urn import guess_entity_type

logger = logging.getLogger(__name__)

_RUN_TABLE_COLUMNS = ["urn", "aspect name", "created at"]
_UNKNOWN_NUM_RECORDS = -1

_DELETE_WITH_REFERENCES_TYPES = {
    "tag",
    "corpuser",
    "corpGroup",
    "domain",
    "glossaryTerm",
    "glossaryNode",
    "form",
}

_RECURSIVE_DELETE_TYPES = {
    "container",
    "dataPlatformInstance",
}


@click.group(cls=DefaultGroup, default="by-filter")
def delete() -> None:
    """Delete metadata from DataHub.

    See `datahub delete by-filter` for the list of available filters.

    See https://datahubproject.io/docs/how/delete-metadata for more detailed docs.
    """
    pass


@dataclass
class DeletionResult:
    num_records: int = 0
    num_timeseries_records: int = 0
    num_entities: int = 0
    num_referenced_entities: int = 0

    def merge(self, another_result: "DeletionResult") -> None:
        self.num_records = self._sum_handle_unknown(
            self.num_records, another_result.num_records
        )
        self.num_timeseries_records = self._sum_handle_unknown(
            self.num_timeseries_records, another_result.num_timeseries_records
        )
        self.num_entities = self._sum_handle_unknown(
            self.num_entities, another_result.num_entities
        )
        self.num_referenced_entities = self._sum_handle_unknown(
            self.num_referenced_entities, another_result.num_referenced_entities
        )

    def format_message(self, *, dry_run: bool, soft: bool, time_sec: float) -> str:
        counters = (
            f"{self.num_entities} entities"
            f" (impacts {self._value_or_unknown(self.num_records)} versioned rows"
            f" and {self._value_or_unknown(self.num_timeseries_records)} timeseries aspect rows)"
        )
        if self.num_referenced_entities > 0:
            counters += (
                f" and cleaned up {self.num_referenced_entities} referenced entities"
            )

        if not dry_run:
            delete_type = "Soft deleted" if soft else "Hard deleted"
            return f"{delete_type} {counters} in {humanfriendly.format_timespan(time_sec)}."
        else:
            return f"[Dry-run] Would delete {counters}."

    @classmethod
    def _value_or_unknown(cls, value: int) -> str:
        return str(value) if value != _UNKNOWN_NUM_RECORDS else "an unknown number of"

    @classmethod
    def _sum_handle_unknown(cls, value1: int, value2: int) -> int:
        if value1 == _UNKNOWN_NUM_RECORDS or value2 == _UNKNOWN_NUM_RECORDS:
            return _UNKNOWN_NUM_RECORDS
        return value1 + value2


@delete.command(no_args_is_help=True)
@click.option(
    "--registry-id", required=True, type=str, help="e.g. mycompany-dq-model:0.0.1"
)
@click.option(
    "--soft/--hard",
    required=False,
    is_flag=True,
    default=True,
    help="specifies soft/hard deletion",
)
@click.option("-n", "--dry-run", required=False, is_flag=True)
@telemetry.with_telemetry()
def by_registry(
    registry_id: str,
    soft: bool,
    dry_run: bool,
) -> None:
    """
    Delete all metadata written using the given registry id and version pair.
    """

    client = get_default_graph()

    if soft and not dry_run:
        raise click.UsageError(
            "Soft-deleting with a registry-id is not yet supported. Try --dry-run to see what you will be deleting, before issuing a hard-delete using the --hard flag"
        )

    with PerfTimer() as timer:
        registry_delete = {"registryId": registry_id, "dryRun": dry_run, "soft": soft}
        (
            structured_rows,
            entities_affected,
            aspects_affected,
            unsafe_aspects,
            unsafe_entity_count,
            unsafe_entities,
        ) = cli_utils.post_rollback_endpoint(
            client._session,
            client.config.server,
            registry_delete,
            "/entities?action=deleteAll",
        )

    if not dry_run:
        message = "soft delete" if soft else "hard delete"
        click.echo(
            f"Took {timer.elapsed_seconds()} seconds to {message}"
            f" {aspects_affected} versioned rows"
            f" for {entities_affected} entities."
        )
    else:
        click.echo(
            f"{entities_affected} entities with {aspects_affected} rows will be affected. "
            f"Took {timer.elapsed_seconds()} seconds to evaluate."
        )
    if structured_rows:
        click.echo(tabulate(structured_rows, _RUN_TABLE_COLUMNS, tablefmt="grid"))


@delete.command(no_args_is_help=True)
@click.option("--urn", required=True, type=str, help="the urn of the entity")
@click.option("-n", "--dry-run", required=False, is_flag=True)
@click.option(
    "-f", "--force", required=False, is_flag=True, help="force the delete if set"
)
@telemetry.with_telemetry()
def references(urn: str, dry_run: bool, force: bool) -> None:
    """
    Delete all references to an entity (but not the entity itself).
    """

    graph = get_default_graph()
    logger.info(f"Using graph: {graph}")

    references_count, related_aspects = graph.delete_references_to_urn(
        urn=urn,
        dry_run=True,
    )

    if references_count == 0:
        click.echo(f"No references to {urn} found")
        return

    click.echo(f"Found {references_count} references to {urn}")
    sample_msg = (
        "\nSample of references\n"
        + tabulate(
            [x.values() for x in related_aspects],
            ["relationship", "entity", "aspect"],
        )
        + "\n"
    )
    click.echo(sample_msg)

    if dry_run:
        logger.info(f"[Dry-run] Would remove {references_count} references to {urn}")
    else:
        if not force:
            click.confirm(
                f"This will delete {references_count} references to {urn} from DataHub. Do you want to continue?",
                abort=True,
            )

        references_count, _ = graph.delete_references_to_urn(
            urn=urn,
            dry_run=False,
        )
        logger.info(f"Deleted {references_count} references to {urn}")


@delete.command()
@click.option("--urn", required=True, type=str, help="the urn of the entity")
def undo_by_filter(urn: str) -> None:
    """
    Undo a soft deletion of an entity
    """
    graph = get_default_graph()
    logger.info(f"Using {graph}")
    graph.set_soft_delete_status(urn=urn, delete=False)


@delete.command(no_args_is_help=True)
@click.option(
    "--urn",
    required=False,
    type=str,
    help="Urn of the entity to delete, for single entity deletion",
)
@click.option(
    "-a",
    "--aspect",
    # This option is inconsistent with rest of CLI but kept for backward compatibility
    "--aspect-name",
    required=False,
    type=str,
    help="The aspect to delete for specified entity(s)",
)
@click.option(
    "-f",
    "--force",
    required=False,
    is_flag=True,
    help="Force deletion, with no confirmation prompts",
)
@click.option(
    "--soft/--hard",
    required=False,
    is_flag=True,
    default=True,
    help="Soft deletion can be undone, while hard deletion removes from the database",
)
@click.option(
    "-e", "--env", required=False, type=str, help="Environment filter (e.g. PROD)"
)
@click.option(
    "-p",
    "--platform",
    required=False,
    type=str,
    help="Platform filter (e.g. snowflake)",
)
@click.option(
    "--entity-type",
    required=False,
    type=str,
    help="Entity type filter (e.g. dataset)",
)
@click.option("--query", required=False, type=str, help="Elasticsearch query string")
@click.option(
    "--recursive",
    required=False,
    is_flag=True,
    help="Recursively delete all contained entities (only for containers and dataPlatformInstances)",
)
@click.option(
    "--start-time",
    required=False,
    type=ClickDatetime(),
    help="Start time (only for timeseries aspects)",
)
@click.option(
    "--end-time",
    required=False,
    type=ClickDatetime(),
    help="End time (only for timeseries aspects)",
)
@click.option(
    "-b",
    "--batch-size",
    required=False,
    default=3000,
    type=int,
    help="Batch size when querying for entities to delete."
    "Maximum 10000. Large batch sizes may cause timeouts.",
)
@click.option(
    "-n",
    "--dry-run",
    required=False,
    is_flag=True,
    help="Print entities to be deleted; perform no deletion",
)
@click.option(
    "--only-soft-deleted",
    required=False,
    is_flag=True,
    default=False,
    help="Only delete soft-deleted entities, for hard deletion",
)
@upgrade.check_upgrade
@telemetry.with_telemetry()
def by_filter(
    urn: Optional[str],
    aspect: Optional[str],
    force: bool,
    soft: bool,
    env: Optional[str],
    platform: Optional[str],
    entity_type: Optional[str],
    query: Optional[str],
    recursive: bool,
    start_time: Optional[datetime],
    end_time: Optional[datetime],
    batch_size: int,
    dry_run: bool,
    only_soft_deleted: bool,
) -> None:
    """Delete metadata from datahub using a single urn or a combination of filters."""

    # Validate the cli arguments.
    _validate_user_urn_and_filters(
        urn=urn,
        entity_type=entity_type,
        platform=platform,
        env=env,
        query=query,
        recursive=recursive,
    )
    soft_delete_filter = _validate_user_soft_delete_flags(
        soft=soft, aspect=aspect, only_soft_deleted=only_soft_deleted
    )
    _validate_user_aspect_flags(aspect=aspect, start_time=start_time, end_time=end_time)
    _validate_batch_size(batch_size)
    # TODO: add some validation on entity_type

    if not force and not soft and not dry_run:
        if only_soft_deleted:
            click.confirm(
                "This will permanently delete data from DataHub. Do you want to continue?",
                abort=True,
            )
        else:
            click.confirm(
                "Hard deletion will permanently delete data from DataHub and can be slow. "
                "We generally recommend using soft deletes instead. "
                "Do you want to continue?",
                abort=True,
            )

    graph = get_default_graph()
    logger.info(f"Using {graph}")

    # Determine which urns to delete.
    delete_by_urn = bool(urn) and not recursive
    if urn:
        urns = [urn]

        if recursive:
            # Add children urns to the list.
            if guess_entity_type(urn) == "dataPlatformInstance":
                urns.extend(
                    graph.get_urns_by_filter(
                        platform_instance=urn,
                        status=soft_delete_filter,
                        batch_size=batch_size,
                    )
                )
            else:
                urns.extend(
                    graph.get_urns_by_filter(
                        container=urn,
                        status=soft_delete_filter,
                        batch_size=batch_size,
                    )
                )
    else:
        urns = list(
            graph.get_urns_by_filter(
                entity_types=[entity_type] if entity_type else None,
                platform=platform,
                env=env,
                query=query,
                status=soft_delete_filter,
                batch_size=batch_size,
            )
        )
        if len(urns) == 0:
            click.echo(
                "Found no urns to delete. Maybe you want to change your filters to be something different?"
            )
            return

    # Print out a summary of the urns to be deleted and confirm with the user.
    if not delete_by_urn:
        urns_by_type: Dict[str, List[str]] = {}
        for urn in urns:
            entity_type = guess_entity_type(urn)
            urns_by_type.setdefault(entity_type, []).append(urn)
        if len(urns_by_type) > 1:
            # Display a breakdown of urns by entity type if there's multiple.
            click.echo("Found urns of multiple entity types")
            for entity_type, entity_urns in urns_by_type.items():
                click.echo(
                    f"- {len(entity_urns)} {entity_type} urn(s). Sample: {choices(entity_urns, k=min(5, len(entity_urns)))}"
                )
        else:
            click.echo(
                f"Found {len(urns)} {entity_type} urn(s). Sample: {choices(urns, k=min(5, len(urns)))}"
            )

        if not force and not dry_run:
            click.confirm(
                f"This will delete {len(urns)} entities from DataHub. Do you want to continue?",
                abort=True,
            )

    urns_iter = urns
    if not delete_by_urn and not dry_run:
        urns_iter = progressbar.progressbar(urns, redirect_stdout=True)

    # Run the deletion.
    deletion_result = DeletionResult()
    with PerfTimer() as timer:
        for urn in urns_iter:
            one_result = _delete_one_urn(
                graph=graph,
                urn=urn,
                aspect_name=aspect,
                soft=soft,
                dry_run=dry_run,
                start_time=start_time,
                end_time=end_time,
            )
            deletion_result.merge(one_result)

    # Report out a summary of the deletion result.
    click.echo(
        deletion_result.format_message(
            dry_run=dry_run, soft=soft, time_sec=timer.elapsed_seconds()
        )
    )


def _validate_user_urn_and_filters(
    urn: Optional[str],
    entity_type: Optional[str],
    platform: Optional[str],
    env: Optional[str],
    query: Optional[str],
    recursive: bool,
) -> None:
    # Check urn / filters options.
    if urn:
        if entity_type or platform or env or query:
            raise click.UsageError(
                "You cannot provide both an urn and a filter rule (entity-type / platform / env / query)."
            )
    elif not urn and not (entity_type or platform or env or query):
        raise click.UsageError(
            "You must provide either an urn or at least one filter (entity-type / platform / env / query) in order to delete entities."
        )
    elif query:
        logger.warning(
            "Using --query is an advanced feature and can easily delete unintended entities. Please use with caution."
        )
    elif env and not (platform or entity_type):
        logger.warning(
            f"Using --env without other filters will delete all metadata in the {env} environment. Please use with caution."
        )

    # Check recursive flag.
    if recursive:
        if not urn:
            raise click.UsageError(
                "The --recursive flag can only be used with a single urn."
            )
        elif guess_entity_type(urn) not in _RECURSIVE_DELETE_TYPES:
            raise click.UsageError(
                f"The --recursive flag can only be used with these entity types: {_RECURSIVE_DELETE_TYPES}."
            )
    elif urn and guess_entity_type(urn) in _RECURSIVE_DELETE_TYPES:
        logger.warning(
            f"This will only delete {urn}. Use --recursive to delete all contained entities."
        )


def _validate_user_soft_delete_flags(
    soft: bool, aspect: Optional[str], only_soft_deleted: bool
) -> RemovedStatusFilter:
    # Check soft / hard delete flags.
    # Note: aspect not None ==> hard delete,
    #    but aspect is None ==> could be either soft or hard delete

    if soft:
        if aspect:
            raise click.UsageError(
                "You cannot provide an aspect name when performing a soft delete. Use --hard to perform a hard delete."
            )

        if only_soft_deleted:
            raise click.UsageError(
                "You cannot provide --only-soft-deleted when performing a soft delete. Use --hard to perform a hard delete."
            )

        soft_delete_filter = RemovedStatusFilter.NOT_SOFT_DELETED
    else:
        # For hard deletes, we will always include the soft-deleted entities, and
        # can optionally filter to exclude non-soft-deleted entities.
        if only_soft_deleted:
            soft_delete_filter = RemovedStatusFilter.ONLY_SOFT_DELETED
        else:
            soft_delete_filter = RemovedStatusFilter.ALL

    return soft_delete_filter


def _validate_user_aspect_flags(
    aspect: Optional[str],
    start_time: Optional[datetime],
    end_time: Optional[datetime],
) -> None:
    # Check the aspect name.
    if aspect and aspect not in ASPECT_MAP:
        logger.info(f"Supported aspects: {list(sorted(ASPECT_MAP.keys()))}")
        raise click.UsageError(
            f"Unknown aspect {aspect}. Ensure the aspect is in the above list."
        )

    # Check that start/end time are set if and only if the aspect is a timeseries aspect.
    if aspect and aspect in TIMESERIES_ASPECT_MAP:
        if not start_time or not end_time:
            raise click.UsageError(
                "You must provide both --start-time and --end-time when deleting a timeseries aspect."
            )
    elif start_time or end_time:
        raise click.UsageError(
            "You can only provide --start-time and --end-time when deleting a timeseries aspect."
        )
    elif aspect:
        raise click.UsageError(
            "Aspect-specific deletion is only supported for timeseries aspects. Please delete the full entity or use a rollback instead."
        )


def _validate_batch_size(batch_size: int) -> None:
    if batch_size <= 0:
        raise click.UsageError("Batch size must be a positive integer.")
    elif batch_size > 10000:
        raise click.UsageError("Batch size cannot exceed 10,000.")


def _delete_one_urn(
    graph: DataHubGraph,
    urn: str,
    soft: bool = False,
    dry_run: bool = False,
    aspect_name: Optional[str] = None,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    run_id: str = "__datahub-delete-cli",
) -> DeletionResult:
    rows_affected: int = 0
    ts_rows_affected: int = 0
    referenced_entities_affected: int = 0

    if soft:
        # Soft delete of entity.
        assert not aspect_name, "aspects cannot be soft deleted"

        if not dry_run:
            graph.soft_delete_entity(urn=urn, run_id=run_id)
        else:
            logger.info(f"[Dry-run] Would soft-delete {urn}")

        rows_affected = 1
        ts_rows_affected = 0

    elif aspect_name and aspect_name in TIMESERIES_ASPECT_MAP:
        # Hard delete of timeseries aspect.

        if not dry_run:
            ts_rows_affected = graph.hard_delete_timeseries_aspect(
                urn=urn,
                aspect_name=aspect_name,
                start_time=start_time,
                end_time=end_time,
            )
        else:
            logger.info(
                f"[Dry-run] Would hard-delete {urn} timeseries aspect {aspect_name}"
            )
            ts_rows_affected = _UNKNOWN_NUM_RECORDS

    elif aspect_name:
        # Hard delete of non-timeseries aspect.

        # TODO: The backend doesn't support this yet.
        raise NotImplementedError(
            "Delete by aspect is not supported yet for non-timeseries aspects. Please delete the full entity or use rollback instead."
        )

    else:
        # Full entity hard delete.
        assert not soft and not aspect_name

        if not dry_run:
            rows_affected, ts_rows_affected = graph.hard_delete_entity(
                urn=urn,
            )
        else:
            logger.info(f"[Dry-run] Would hard-delete {urn}")
            rows_affected = _UNKNOWN_NUM_RECORDS
            ts_rows_affected = _UNKNOWN_NUM_RECORDS

        # For full entity deletes, we also might clean up references to the entity.
        if guess_entity_type(urn) in _DELETE_WITH_REFERENCES_TYPES:
            referenced_entities_affected, _ = graph.delete_references_to_urn(
                urn=urn,
                dry_run=dry_run,
            )
            if dry_run and referenced_entities_affected > 0:
                logger.info(
                    f"[Dry-run] Would remove {referenced_entities_affected} references to {urn}"
                )

    return DeletionResult(
        num_entities=1,
        num_records=rows_affected,
        num_timeseries_records=ts_rows_affected,
        num_referenced_entities=referenced_entities_affected,
    )



import contextlib
import logging
import subprocess
from typing import Callable, Iterator, List, Optional, Union

import pytest
import pytest_docker.plugin

logger = logging.getLogger(__name__)


def is_responsive(container_name: str, port: int, hostname: Optional[str]) -> bool:
    """A cheap way to figure out if a port is responsive on a container"""
    if hostname:
        cmd = f"docker exec {container_name} /bin/bash -c 'echo -n > /dev/tcp/{hostname}/{port}'"
    else:
        # use the hostname of the container
        cmd = f"docker exec {container_name} /bin/bash -c 'c_host=`hostname`;echo -n > /dev/tcp/$c_host/{port}'"
    ret = subprocess.run(
        cmd,
        shell=True,
    )
    return ret.returncode == 0


def wait_for_port(
    docker_services: pytest_docker.plugin.Services,
    container_name: str,
    container_port: int,
    hostname: Optional[str] = None,
    timeout: float = 30.0,
    pause: float = 0.5,
    checker: Optional[Callable[[], bool]] = None,
) -> None:
    try:
        docker_services.wait_until_responsive(
            timeout=timeout,
            pause=pause,
            check=(
                checker
                if checker
                else lambda: is_responsive(container_name, container_port, hostname)
            ),
        )
        logger.info(f"Container {container_name} is ready!")
    finally:
        # use check=True to raise an error if command gave bad exit code
        subprocess.run(f"docker logs {container_name}", shell=True, check=True)


@pytest.fixture(scope="module")
def docker_compose_runner(
    docker_compose_command, docker_compose_project_name, docker_setup, docker_cleanup
):
    @contextlib.contextmanager
    def run(
        compose_file_path: Union[str, List[str]], key: str, cleanup: bool = True
    ) -> Iterator[pytest_docker.plugin.Services]:
        with pytest_docker.plugin.get_docker_services(
            docker_compose_command=docker_compose_command,
            # We can remove the type ignore once this is merged:
            # https://github.com/avast/pytest-docker/pull/108
            docker_compose_file=compose_file_path,  # type: ignore
            docker_compose_project_name=f"{docker_compose_project_name}-{key}",
            docker_setup=docker_setup,
            docker_cleanup=docker_cleanup if cleanup else [],
        ) as docker_services:
            yield docker_services

    return run



import json
import logging
from pathlib import Path

import click
from click_default_group import DefaultGroup

from datahub.api.entities.structuredproperties.structuredproperties import (
    StructuredProperties,
)
from datahub.ingestion.graph.client import get_default_graph
from datahub.telemetry import telemetry
from datahub.upgrade import upgrade
from datahub.utilities.urns.urn import Urn

logger = logging.getLogger(__name__)


@click.group(cls=DefaultGroup, default="upsert")
def properties() -> None:
    """A group of commands to interact with structured properties in DataHub."""
    pass


@properties.command(
    name="upsert",
)
@click.option("-f", "--file", required=True, type=click.Path(exists=True))
@upgrade.check_upgrade
@telemetry.with_telemetry()
def upsert(file: Path) -> None:
    """Upsert structured properties in DataHub."""

    StructuredProperties.create(str(file))


@properties.command(
    name="get",
)
@click.option("--urn", required=True, type=str)
@click.option("--to-file", required=False, type=str)
@upgrade.check_upgrade
@telemetry.with_telemetry()
def get(urn: str, to_file: str) -> None:
    """Get structured properties from DataHub"""
    urn = Urn.make_structured_property_urn(urn)

    with get_default_graph() as graph:
        if graph.exists(urn):
            structuredproperties: StructuredProperties = (
                StructuredProperties.from_datahub(graph=graph, urn=urn)
            )
            click.secho(
                f"{json.dumps(structuredproperties.dict(exclude_unset=True, exclude_none=True), indent=2)}"
            )
            if to_file:
                structuredproperties.to_yaml(Path(to_file))
                click.secho(
                    f"Structured property yaml written to {to_file}", fg="green"
                )
        else:
            click.secho(f"Structured property {urn} does not exist")



import glob
import logging
import pathlib
from dataclasses import dataclass
from typing import Dict, List, Optional, Set

from datahub.ingestion.source.looker.looker_connection import LookerConnectionDefinition
from datahub.ingestion.source.looker.looker_template_language import (
    load_and_preprocess_file,
)
from datahub.ingestion.source.looker.lookml_config import (
    BASE_PROJECT_NAME,
    EXPLORE_FILE_EXTENSION,
    LookMLSourceConfig,
    LookMLSourceReport,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True, order=True)
class ProjectInclude:
    project: str
    include: str


@dataclass
class LookerField:
    name: str
    primary_key: str  # possible values yes and no
    type: str
    sql: Optional[str]


@dataclass
class LookerModel:
    connection: str
    includes: List[str]
    explores: List[dict]
    resolved_includes: List[ProjectInclude]

    @staticmethod
    def from_looker_dict(
        looker_model_dict: dict,
        base_project_name: str,
        root_project_name: Optional[str],
        base_projects_folders: Dict[str, pathlib.Path],
        path: str,
        source_config: LookMLSourceConfig,
        reporter: LookMLSourceReport,
    ) -> "LookerModel":
        logger.debug(f"Loading model from {path}")
        connection = looker_model_dict["connection"]
        includes = looker_model_dict.get("includes", [])
        resolved_includes = LookerModel.resolve_includes(
            includes,
            base_project_name,
            root_project_name,
            base_projects_folders,
            path,
            source_config,
            reporter,
            seen_so_far=set(),
            traversal_path=pathlib.Path(path).stem,
        )
        logger.debug(f"{path} has resolved_includes: {resolved_includes}")
        explores = looker_model_dict.get("explores", [])

        explore_files = [
            x.include
            for x in resolved_includes
            if x.include.endswith(EXPLORE_FILE_EXTENSION)
        ]
        for included_file in explore_files:
            try:
                parsed = load_and_preprocess_file(
                    path=included_file,
                    source_config=source_config,
                )
                included_explores = parsed.get("explores", [])
                explores.extend(included_explores)
            except Exception as e:
                reporter.report_warning(
                    title="Error Loading Include",
                    message="Failed to load include file",
                    context=f"Include Details: {included_file}",
                    exc=e,
                )
                # continue in this case, as it might be better to load and resolve whatever we can

        return LookerModel(
            connection=connection,
            includes=includes,
            resolved_includes=resolved_includes,
            explores=explores,
        )

    @staticmethod
    def resolve_includes(
        includes: List[str],
        project_name: str,
        root_project_name: Optional[str],
        base_projects_folder: Dict[str, pathlib.Path],
        path: str,
        source_config: LookMLSourceConfig,
        reporter: LookMLSourceReport,
        seen_so_far: Set[str],
        traversal_path: str = "",  # a cosmetic parameter to aid debugging
    ) -> List[ProjectInclude]:
        """Resolve ``include`` statements in LookML model files to a list of ``.lkml`` files.

        For rules on how LookML ``include`` statements are written, see
            https://docs.looker.com/data-modeling/getting-started/ide-folders#wildcard_examples
        """

        resolved = []
        for inc in includes:
            # Filter out dashboards - we get those through the looker source.
            if (
                inc.endswith(".dashboard")
                or inc.endswith(".dashboard.lookml")
                or inc.endswith(".dashboard.lkml")
            ):
                logger.debug(f"include '{inc}' is a dashboard, skipping it")
                continue

            resolved_project_name = project_name
            resolved_project_folder = str(base_projects_folder[project_name])

            # Massage the looker include into a valid glob wildcard expression
            if inc.startswith("//"):
                # remote include, let's see if we have the project checked out locally
                (remote_project, project_local_path) = inc[2:].split("/", maxsplit=1)
                if remote_project in base_projects_folder:
                    resolved_project_folder = str(base_projects_folder[remote_project])
                    glob_expr = f"{resolved_project_folder}/{project_local_path}"
                    resolved_project_name = remote_project
                else:
                    logger.warning(
                        f"Resolving {inc} failed. Could not find a locally checked out reference for {remote_project}"
                    )
                    continue
            elif inc.startswith("/"):
                glob_expr = f"{resolved_project_folder}{inc}"

                # The include path is sometimes '/{project_name}/{path_within_project}'
                # instead of '//{project_name}/{path_within_project}' or '/{path_within_project}'.
                #
                # TODO: I can't seem to find any documentation on this pattern, but we definitely
                # have seen it in the wild. Example from Mozilla's public looker-hub repo:
                # https://github.com/mozilla/looker-hub/blob/f491ca51ce1add87c338e6723fd49bc6ae4015ca/fenix/explores/activation.explore.lkml#L7
                # As such, we try to handle it but are as defensive as possible.

                non_base_project_name = project_name
                if project_name == BASE_PROJECT_NAME and root_project_name is not None:
                    non_base_project_name = root_project_name
                if non_base_project_name != BASE_PROJECT_NAME and inc.startswith(
                    f"/{non_base_project_name}/"
                ):
                    # This might be a local include. Let's make sure that '/{project_name}' doesn't
                    # exist as normal include in the project.
                    if not pathlib.Path(
                        f"{resolved_project_folder}/{non_base_project_name}"
                    ).exists():
                        path_within_project = pathlib.Path(*pathlib.Path(inc).parts[2:])
                        glob_expr = f"{resolved_project_folder}/{path_within_project}"
            else:
                # Need to handle a relative path.
                glob_expr = str(pathlib.Path(path).parent / inc)
            # "**" matches an arbitrary number of directories in LookML
            # we also resolve these paths to absolute paths so we can de-dup effectively later on
            included_files = [
                str(p.resolve())
                for p in [
                    pathlib.Path(p)
                    for p in sorted(
                        glob.glob(glob_expr, recursive=True)
                        + glob.glob(f"{glob_expr}.lkml", recursive=True)
                    )
                ]
                # We don't want to match directories. The '**' glob can be used to
                # recurse into directories.
                if p.is_file()
            ]
            logger.debug(
                f"traversal_path={traversal_path}, included_files = {included_files}, seen_so_far: {seen_so_far}"
            )
            if "*" not in inc and not included_files:
                reporter.report_failure(
                    title="Error Resolving Include",
                    message=f"Cannot resolve include {inc}",
                    context=f"Path: {path}",
                )
            elif not included_files:
                reporter.report_failure(
                    title="Error Resolving Include",
                    message=f"Did not resolve anything for wildcard include {inc}",
                    context=f"Path: {path}",
                )
            # only load files that we haven't seen so far
            included_files = [x for x in included_files if x not in seen_so_far]
            for included_file in included_files:
                # Filter out dashboards - we get those through the looker source.
                if (
                    included_file.endswith(".dashboard")
                    or included_file.endswith(".dashboard.lookml")
                    or included_file.endswith(".dashboard.lkml")
                ):
                    logger.debug(
                        f"include '{included_file}' is a dashboard, skipping it"
                    )
                    continue

                logger.debug(
                    f"Will be loading {included_file}, traversed here via {traversal_path}"
                )
                try:
                    parsed = load_and_preprocess_file(
                        path=included_file,
                        source_config=source_config,
                    )
                    seen_so_far.add(included_file)
                    if "includes" in parsed:  # we have more includes to resolve!
                        resolved.extend(
                            LookerModel.resolve_includes(
                                parsed["includes"],
                                resolved_project_name,
                                root_project_name,
                                base_projects_folder,
                                included_file,
                                source_config,
                                reporter,
                                seen_so_far,
                                traversal_path=traversal_path
                                + "."
                                + pathlib.Path(included_file).stem,
                            )
                        )
                except Exception as e:
                    reporter.report_warning(
                        title="Error Loading Include File",
                        message="Failed to load included file",
                        context=f"Include Details: {included_file}",
                        exc=e,
                    )
                    # continue in this case, as it might be better to load and resolve whatever we can

            resolved.extend(
                [
                    ProjectInclude(project=resolved_project_name, include=f)
                    for f in included_files
                ]
            )
        return resolved


@dataclass
class LookerViewFile:
    absolute_file_path: str
    connection: Optional[LookerConnectionDefinition]
    includes: List[str]
    resolved_includes: List[ProjectInclude]
    views: List[Dict]
    raw_file_content: str

    @classmethod
    def from_looker_dict(
        cls,
        absolute_file_path: str,
        looker_view_file_dict: dict,
        project_name: str,
        root_project_name: Optional[str],
        base_projects_folder: Dict[str, pathlib.Path],
        raw_file_content: str,
        source_config: LookMLSourceConfig,
        reporter: LookMLSourceReport,
    ) -> "LookerViewFile":
        logger.debug(f"Loading view file at {absolute_file_path}")
        includes = looker_view_file_dict.get("includes", [])
        resolved_path = str(pathlib.Path(absolute_file_path).resolve())
        seen_so_far = set()
        seen_so_far.add(resolved_path)
        resolved_includes = LookerModel.resolve_includes(
            includes,
            project_name,
            root_project_name,
            base_projects_folder,
            absolute_file_path,
            source_config,
            reporter,
            seen_so_far=seen_so_far,
        )
        logger.debug(
            f"resolved_includes for {absolute_file_path} is {resolved_includes}"
        )
        views = looker_view_file_dict.get("views", [])

        return cls(
            absolute_file_path=absolute_file_path,
            connection=None,
            includes=includes,
            resolved_includes=resolved_includes,
            views=views,
            raw_file_content=raw_file_content,
        )



import os
from typing import List, Literal, Optional

import certifi
from pydantic import Field, validator

from datahub.configuration.common import AllowDenyPattern, ConfigModel
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.ingestion.source.ge_profiling_config import GEProfilingBaseConfig
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)
from datahub.ingestion.source.usage.usage_common import BaseUsageConfig
from datahub.ingestion.source_config.operation_config import is_profiling_enabled


class DremioConnectionConfig(ConfigModel):
    hostname: Optional[str] = Field(
        default=None,
        description="Hostname or IP Address of the Dremio server",
    )

    port: int = Field(
        default=9047,
        description="Port of the Dremio REST API",
    )

    username: Optional[str] = Field(
        default=None,
        description="Dremio username",
    )

    authentication_method: Optional[str] = Field(
        default="PAT",
        description="Authentication method: 'password' or 'PAT' (Personal Access Token)",
    )

    password: Optional[str] = Field(
        default=None,
        description="Dremio password or Personal Access Token",
    )

    tls: bool = Field(
        default=True,
        description="Whether the Dremio REST API port is encrypted",
    )

    disable_certificate_verification: Optional[bool] = Field(
        default=False,
        description="Disable TLS certificate verification",
    )

    path_to_certificates: str = Field(
        default=certifi.where(),
        description="Path to SSL certificates",
    )

    is_dremio_cloud: bool = Field(
        default=False,
        description="Whether this is a Dremio Cloud instance",
    )

    # https://docs.dremio.com/cloud/reference/api/#api-endpoint
    dremio_cloud_region: Literal["US", "EU"] = Field(
        default="US",
        description="Dremio Cloud region ('US' or 'EU')",
    )

    dremio_cloud_project_id: Optional[str] = Field(
        default=None,
        description="ID of Dremio Cloud Project. Found in Project Settings in the Dremio Cloud UI",
    )

    @validator("authentication_method")
    def validate_auth_method(cls, value):
        allowed_methods = ["password", "PAT"]
        if value not in allowed_methods:
            raise ValueError(
                f"authentication_method must be one of {allowed_methods}",
            )
        return value

    @validator("password")
    def validate_password(cls, value, values):
        if values.get("authentication_method") == "PAT" and not value:
            raise ValueError(
                "Password (Personal Access Token) is required when using PAT authentication",
            )
        return value


class ProfileConfig(GEProfilingBaseConfig):
    query_timeout: int = Field(
        default=300, description="Time before cancelling Dremio profiling query"
    )
    include_field_median_value: bool = Field(
        default=False,
        hidden_from_docs=True,
        description="Median causes a number of issues in Dremio.",
    )


class DremioSourceMapping(EnvConfigMixin, PlatformInstanceConfigMixin, ConfigModel):
    platform: str = Field(
        description="Source connection made by Dremio (e.g. S3, Snowflake)",
    )
    source_name: str = Field(
        description="Alias of platform in Dremio connection",
    )


class DremioSourceConfig(
    DremioConnectionConfig,
    StatefulIngestionConfigBase,
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
):
    domain: Optional[str] = Field(
        default=None,
        description="Domain for all source objects.",
    )

    source_mappings: Optional[List[DremioSourceMapping]] = Field(
        default=None,
        description="Mappings from Dremio sources to DataHub platforms and datasets.",
    )

    # Entity Filters
    schema_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for schemas to filter",
    )

    dataset_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for tables and views to filter in ingestion. Specify regex to match the entire table name in dremio.schema.table format. e.g. to match all tables starting with customer in Customer database and public schema, use the regex 'dremio.public.customer.*'",
    )

    usage: BaseUsageConfig = Field(
        description="The usage config to use when generating usage statistics",
        default=BaseUsageConfig(),
    )

    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = None

    # Profiling
    profile_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for tables to profile",
    )
    profiling: ProfileConfig = Field(
        default=ProfileConfig(),
        description="Configuration for profiling",
    )

    def is_profiling_enabled(self) -> bool:
        return self.profiling.enabled and is_profiling_enabled(
            self.profiling.operation_config
        )

    # Advanced Configs
    max_workers: int = Field(
        default=5 * (os.cpu_count() or 4),
        description="Number of worker threads to use for parallel processing",
    )

    include_query_lineage: bool = Field(
        default=False,
        description="Whether to include query-based lineage information.",
    )



from dataclasses import dataclass, field
from typing import Optional, Tuple

from datahub.ingestion.api.report import EntityFilterReport, Report
from datahub.ingestion.source.sql.sql_report import SQLSourceReport
from datahub.ingestion.source_report.ingestion_stage import IngestionStageReport
from datahub.utilities.lossy_collections import LossyDict, LossyList
from datahub.utilities.perf_timer import PerfTimer


@dataclass
class UnityCatalogUsagePerfReport(Report):
    get_queries_timer: PerfTimer = field(default_factory=PerfTimer)
    sql_parsing_timer: PerfTimer = field(default_factory=PerfTimer)
    spark_sql_parsing_timer: PerfTimer = field(default_factory=PerfTimer)
    aggregator_add_event_timer: PerfTimer = field(default_factory=PerfTimer)
    gen_operation_timer: PerfTimer = field(default_factory=PerfTimer)
    query_fingerprinting_timer: PerfTimer = field(default_factory=PerfTimer)


@dataclass
class UnityCatalogReport(IngestionStageReport, SQLSourceReport):
    metastores: EntityFilterReport = EntityFilterReport.field(type="metastore")
    catalogs: EntityFilterReport = EntityFilterReport.field(type="catalog")
    schemas: EntityFilterReport = EntityFilterReport.field(type="schema")
    tables: EntityFilterReport = EntityFilterReport.field(type="table/view")
    table_profiles: EntityFilterReport = EntityFilterReport.field(type="table profile")
    notebooks: EntityFilterReport = EntityFilterReport.field(type="notebook")

    hive_metastore_catalog_found: Optional[bool] = None

    num_column_lineage_skipped_column_count: int = 0
    num_external_upstreams_lacking_permissions: int = 0
    num_external_upstreams_unsupported: int = 0

    num_queries: int = 0
    num_unique_queries: int = 0
    num_queries_dropped_parse_failure: int = 0
    num_queries_missing_table: int = 0  # Can be due to pattern filter
    num_queries_duplicate_table: int = 0
    num_queries_parsed_by_spark_plan: int = 0
    usage_perf_report: UnityCatalogUsagePerfReport = field(
        default_factory=UnityCatalogUsagePerfReport
    )

    # Distinguish from Operations emitted for created / updated timestamps
    num_operational_stats_workunits_emitted: int = 0

    profile_table_timeouts: LossyList[str] = field(default_factory=LossyList)
    profile_table_empty: LossyList[str] = field(default_factory=LossyList)
    profile_table_errors: LossyDict[str, LossyList[Tuple[str, str]]] = field(
        default_factory=LossyDict
    )
    num_profile_missing_size_in_bytes: int = 0
    num_profile_failed_unsupported_column_type: int = 0
    num_profile_failed_int_casts: int = 0

    num_catalogs_missing_name: int = 0
    num_schemas_missing_name: int = 0
    num_tables_missing_name: int = 0
    num_columns_missing_name: int = 0
    num_queries_missing_info: int = 0



from typing import Dict

from typing_extensions import Final

from datahub.metadata.schema_classes import JobStatusClass


class SageMakerJobInfo:
    # Note: The weird typing for the following commands is because the child classes
    # use a Final type to ensure that they're interpreted by the boto3 stubs correctly.
    # If we try to type these as plain strings, we get a TypeError because it's being converted
    # to a non-overwritable type.
    # See https://mypy.readthedocs.io/en/stable/final_attrs.html#details-of-using-final

    # boto3 command to get list of jobs
    @property
    def list_command(self) -> str:
        raise NotImplementedError

    # field in job listing response containing actual list
    @property
    def list_key(self) -> str:
        raise NotImplementedError

    # field in job listing response element corresponding to job name
    @property
    def list_name_key(self) -> str:
        raise NotImplementedError

    # field in job listing response element corresponding to job ARN
    @property
    def list_arn_key(self) -> str:
        raise NotImplementedError

    # boto3 command to get job details
    @property
    def describe_command(self) -> str:
        raise NotImplementedError

    # field in job description response corresponding to job name
    @property
    def describe_name_key(self) -> str:
        raise NotImplementedError

    # field in job description response corresponding to job ARN
    @property
    def describe_arn_key(self) -> str:
        raise NotImplementedError

    # field in job description response corresponding to job status
    @property
    def describe_status_key(self) -> str:
        raise NotImplementedError

    # job-specific mapping from boto3 status strings to DataHub-native enum
    status_map: Dict[str, str]

    # name of function for processing job for ingestion
    processor: str


class AutoMlJobInfo(SageMakerJobInfo):
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.list_auto_ml_jobs
    list_command: Final = "list_auto_ml_jobs"
    list_key: Final = "AutoMLJobSummaries"
    list_name_key: Final = "AutoMLJobName"
    list_arn_key: Final = "AutoMLJobArn"
    # DescribeAutoMLJobV2 are new versions of CreateAutoMLJob and DescribeAutoMLJob which offer backward compatibility.
    # https://docs.aws.amazon.com/sagemaker/latest/APIReference/API_DescribeAutoMLJobV2.html
    describe_command: Final = "describe_auto_ml_job_v2"
    describe_name_key: Final = "AutoMLJobName"
    describe_arn_key: Final = "AutoMLJobArn"
    describe_status_key: Final = "AutoMLJobStatus"
    status_map = {
        "Completed": JobStatusClass.COMPLETED,
        "InProgress": JobStatusClass.IN_PROGRESS,
        "Failed": JobStatusClass.FAILED,
        "Stopped": JobStatusClass.STOPPED,
        "Stopping": JobStatusClass.STOPPING,
    }

    processor = "process_auto_ml_job"


class CompilationJobInfo(SageMakerJobInfo):
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.list_compilation_jobs
    list_command: Final = "list_compilation_jobs"
    list_key: Final = "CompilationJobSummaries"
    list_name_key: Final = "CompilationJobName"
    list_arn_key: Final = "CompilationJobArn"
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.describe_compilation_job
    describe_command: Final = "describe_compilation_job"
    describe_name_key: Final = "CompilationJobName"
    describe_arn_key: Final = "CompilationJobArn"
    describe_status_key: Final = "CompilationJobStatus"
    status_map = {
        "INPROGRESS": JobStatusClass.IN_PROGRESS,
        "COMPLETED": JobStatusClass.COMPLETED,
        "FAILED": JobStatusClass.FAILED,
        "STARTING": JobStatusClass.STARTING,
        "STOPPING": JobStatusClass.STOPPING,
        "STOPPED": JobStatusClass.STOPPED,
    }
    processor = "process_compilation_job"


class HyperParameterTuningJobInfo(SageMakerJobInfo):
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.list_hyper_parameter_tuning_jobs
    list_command: Final = "list_hyper_parameter_tuning_jobs"
    list_key: Final = "HyperParameterTuningJobSummaries"
    list_name_key: Final = "HyperParameterTuningJobName"
    list_arn_key: Final = "HyperParameterTuningJobArn"
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.describe_hyper_parameter_tuning_job
    describe_command: Final = "describe_hyper_parameter_tuning_job"
    describe_name_key: Final = "HyperParameterTuningJobName"
    describe_arn_key: Final = "HyperParameterTuningJobArn"
    describe_status_key: Final = "HyperParameterTuningJobStatus"
    status_map = {
        "InProgress": JobStatusClass.IN_PROGRESS,
        "Completed": JobStatusClass.COMPLETED,
        "Failed": JobStatusClass.FAILED,
        "Stopping": JobStatusClass.STOPPING,
        "Stopped": JobStatusClass.STOPPED,
    }
    processor = "process_hyper_parameter_tuning_job"


class LabelingJobInfo(SageMakerJobInfo):
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.list_labeling_jobs
    list_command: Final = "list_labeling_jobs"
    list_key: Final = "LabelingJobSummaryList"
    list_name_key: Final = "LabelingJobName"
    list_arn_key: Final = "LabelingJobArn"
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.describe_labeling_job
    describe_command: Final = "describe_labeling_job"
    describe_name_key: Final = "LabelingJobName"
    describe_arn_key: Final = "LabelingJobArn"
    describe_status_key: Final = "LabelingJobStatus"
    status_map = {
        "Initializing": JobStatusClass.STARTING,
        "InProgress": JobStatusClass.IN_PROGRESS,
        "Completed": JobStatusClass.COMPLETED,
        "Failed": JobStatusClass.FAILED,
        "Stopping": JobStatusClass.STOPPING,
        "Stopped": JobStatusClass.STOPPED,
    }
    processor = "process_labeling_job"


class ProcessingJobInfo(SageMakerJobInfo):
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.list_processing_jobs
    list_command: Final = "list_processing_jobs"
    list_key: Final = "ProcessingJobSummaries"
    list_name_key: Final = "ProcessingJobName"
    list_arn_key: Final = "ProcessingJobArn"
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.describe_processing_job
    describe_command: Final = "describe_processing_job"
    describe_name_key: Final = "ProcessingJobName"
    describe_arn_key: Final = "ProcessingJobArn"
    describe_status_key: Final = "ProcessingJobStatus"
    status_map = {
        "InProgress": JobStatusClass.IN_PROGRESS,
        "Completed": JobStatusClass.COMPLETED,
        "Failed": JobStatusClass.FAILED,
        "Stopping": JobStatusClass.STOPPING,
        "Stopped": JobStatusClass.STOPPED,
    }
    processor = "process_processing_job"


class TrainingJobInfo(SageMakerJobInfo):
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.list_training_jobs
    list_command: Final = "list_training_jobs"
    list_key: Final = "TrainingJobSummaries"
    list_name_key: Final = "TrainingJobName"
    list_arn_key: Final = "TrainingJobArn"
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.describe_training_job
    describe_command: Final = "describe_training_job"
    describe_name_key: Final = "TrainingJobName"
    describe_arn_key: Final = "TrainingJobArn"
    describe_status_key: Final = "TrainingJobStatus"
    status_map = {
        "InProgress": JobStatusClass.IN_PROGRESS,
        "Completed": JobStatusClass.COMPLETED,
        "Failed": JobStatusClass.FAILED,
        "Stopping": JobStatusClass.STOPPING,
        "Stopped": JobStatusClass.STOPPED,
    }
    processor = "process_training_job"


class TransformJobInfo(SageMakerJobInfo):
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.list_transform_jobs
    list_command: Final = "list_transform_jobs"
    list_key: Final = "TransformJobSummaries"
    list_name_key: Final = "TransformJobName"
    list_arn_key: Final = "TransformJobArn"
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker.html#SageMaker.Client.describe_transform_job
    describe_command: Final = "describe_transform_job"
    describe_name_key: Final = "TransformJobName"
    describe_arn_key: Final = "TransformJobArn"
    describe_status_key: Final = "TransformJobStatus"
    status_map = {
        "InProgress": JobStatusClass.IN_PROGRESS,
        "Completed": JobStatusClass.COMPLETED,
        "Failed": JobStatusClass.FAILED,
        "Stopping": JobStatusClass.STOPPING,
        "Stopped": JobStatusClass.STOPPED,
    }
    processor = "process_transform_job"



import logging
from datetime import datetime
from typing import Dict, Iterable, List, Tuple

from confluent_kafka import (
    OFFSET_BEGINNING,
    Consumer,
    DeserializingConsumer,
    TopicPartition,
)
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import AvroDeserializer

from datahub.configuration.kafka import KafkaConsumerConnectionConfig
from datahub.ingestion.api.closeable import Closeable
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.source.datahub.config import DataHubSourceConfig
from datahub.ingestion.source.datahub.report import DataHubSourceReport
from datahub.ingestion.source.datahub.state import PartitionOffset
from datahub.metadata.schema_classes import MetadataChangeLogClass

logger = logging.getLogger(__name__)

KAFKA_GROUP_PREFIX = "datahub_source"


class DataHubKafkaReader(Closeable):
    def __init__(
        self,
        config: DataHubSourceConfig,
        connection_config: KafkaConsumerConnectionConfig,
        report: DataHubSourceReport,
        ctx: PipelineContext,
    ):
        self.config = config
        self.connection_config = connection_config
        self.report = report
        self.group_id = f"{KAFKA_GROUP_PREFIX}-{ctx.pipeline_name}"
        self.ctx = ctx

    def __enter__(self) -> "DataHubKafkaReader":
        self.consumer = DeserializingConsumer(
            {
                "group.id": self.group_id,
                "bootstrap.servers": self.connection_config.bootstrap,
                **self.connection_config.consumer_config,
                "auto.offset.reset": "earliest",
                "enable.auto.commit": False,
                "value.deserializer": AvroDeserializer(
                    schema_registry_client=SchemaRegistryClient(
                        {"url": self.connection_config.schema_registry_url}
                    ),
                    return_record_name=True,
                ),
            }
        )
        return self

    def get_mcls(
        self, from_offsets: Dict[int, int], stop_time: datetime
    ) -> Iterable[Tuple[MetadataChangeLogClass, PartitionOffset]]:
        # Based on https://github.com/confluentinc/confluent-kafka-python/issues/145#issuecomment-284843254
        def on_assign(consumer: Consumer, partitions: List[TopicPartition]) -> None:
            for p in partitions:
                p.offset = from_offsets.get(p.partition, OFFSET_BEGINNING)
                logger.debug(f"Set partition {p.partition} offset to {p.offset}")
            consumer.assign(partitions)

        self.consumer.subscribe([self.config.kafka_topic_name], on_assign=on_assign)
        try:
            yield from self._poll_partition(stop_time)
        finally:
            self.consumer.unsubscribe()

    def _poll_partition(
        self, stop_time: datetime
    ) -> Iterable[Tuple[MetadataChangeLogClass, PartitionOffset]]:
        while True:
            msg = self.consumer.poll(10)
            if msg is None:
                break

            try:
                mcl = MetadataChangeLogClass.from_obj(msg.value(), True)
            except Exception as e:
                logger.warning(f"Error deserializing MCL: {e}")
                self.report.num_kafka_parse_errors += 1
                self.report.kafka_parse_errors.setdefault(str(e), 0)
                self.report.kafka_parse_errors[str(e)] += 1
                continue

            if mcl.created and mcl.created.time > stop_time.timestamp() * 1000:
                logger.info(
                    f"Stopped reading from kafka, reached MCL "
                    f"with audit stamp {datetime.fromtimestamp(mcl.created.time / 1000)}"
                )
                break

            if mcl.aspectName and mcl.aspectName in self.config.exclude_aspects:
                self.report.num_kafka_excluded_aspects += 1
                continue

            # TODO: Consider storing state in kafka instead, via consumer.commit()
            yield mcl, PartitionOffset(partition=msg.partition(), offset=msg.offset())

    def close(self) -> None:
        self.consumer.close()



import datetime
import json
import logging
import os.path
import pathlib
from dataclasses import dataclass, field
from enum import auto
from functools import partial
from typing import Any, Iterable, Iterator, List, Optional, Tuple, Union

import ijson
from pydantic import validator
from pydantic.fields import Field

from datahub.configuration.common import ConfigEnum
from datahub.configuration.validate_field_deprecation import pydantic_field_deprecated
from datahub.configuration.validate_field_rename import pydantic_renamed_field
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import (
    CapabilityReport,
    MetadataWorkUnitProcessor,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.api.source_helpers import (
    auto_status_aspect,
    auto_workunit_reporter,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.fs.fs_base import FileInfo, get_path_schema
from datahub.ingestion.fs.fs_registry import fs_registry
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
    StaleEntityRemovalSourceReport,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
    StatefulIngestionSourceBase,
)
from datahub.metadata.com.linkedin.pegasus2avro.mxe import (
    MetadataChangeEvent,
    MetadataChangeProposal,
)
from datahub.metadata.schema_classes import UsageAggregationClass

logger = logging.getLogger(__name__)


class FileReadMode(ConfigEnum):
    STREAM = auto()
    BATCH = auto()
    AUTO = auto()


class FileSourceConfig(StatefulIngestionConfigBase):
    _filename = pydantic_field_deprecated(
        "filename",
        message="filename is deprecated. Use path instead.",
    )
    path: str = Field(
        description=(
            "File path to folder or file to ingest, or URL to a remote file. "
            "If pointed to a folder, all files with extension {file_extension} (default json) within that folder will be processed."
        )
    )
    file_extension: str = Field(
        ".json",
        description=(
            "When providing a folder to use to read files, set this field to control file extensions that you want the source to process. "
            "* is a special value that means process every file regardless of extension"
        ),
    )
    read_mode: FileReadMode = FileReadMode.AUTO
    aspect: Optional[str] = Field(
        default=None,
        description="Set to an aspect to only read this aspect for ingestion.",
    )
    count_all_before_starting: bool = Field(
        default=True,
        description=(
            "When enabled, counts total number of records in the file before starting. "
            "Used for accurate estimation of completion time. Turn it off if startup time is too high."
        ),
    )

    _minsize_for_streaming_mode_in_bytes: int = (
        100 * 1000 * 1000  # Must be at least 100MB before we use streaming mode
    )

    _filename_populates_path_if_present = pydantic_renamed_field(
        "filename", "path", print_warning=False
    )

    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = None

    @validator("file_extension", always=True)
    def add_leading_dot_to_extension(cls, v: str) -> str:
        if v:
            if v.startswith("."):
                return v
            else:
                return "." + v
        return v


@dataclass
class FileSourceReport(StaleEntityRemovalSourceReport):
    total_num_files: int = 0
    num_files_completed: int = 0
    files_completed: list = field(default_factory=list)
    percentage_completion: str = "0%"
    estimated_time_to_completion_in_minutes: int = -1
    total_bytes_on_disk: Optional[int] = None
    total_bytes_read_completed_files: int = 0
    current_file_name: Optional[str] = None
    current_file_size: Optional[int] = None
    current_file_num_elements: Optional[int] = None
    current_file_bytes_read: Optional[int] = None
    current_file_elements_read: Optional[int] = None
    total_parse_time_in_seconds: float = 0
    total_count_time_in_seconds: float = 0
    total_deserialize_time_in_seconds: float = 0

    def add_deserialize_time(self, delta: datetime.timedelta) -> None:
        self.total_deserialize_time_in_seconds += round(delta.total_seconds(), 2)

    def add_parse_time(self, delta: datetime.timedelta) -> None:
        self.total_parse_time_in_seconds += round(delta.total_seconds(), 2)

    def add_count_time(self, delta: datetime.timedelta) -> None:
        self.total_count_time_in_seconds += round(delta.total_seconds(), 2)

    def append_total_bytes_on_disk(self, delta: int) -> None:
        if self.total_bytes_on_disk is not None:
            self.total_bytes_on_disk += delta
        else:
            self.total_bytes_on_disk = delta

    def reset_current_file_stats(self) -> None:
        self.current_file_name = None
        self.current_file_bytes_read = None
        self.current_file_num_elements = None
        self.current_file_elements_read = None

    def compute_stats(self) -> None:
        super().compute_stats()
        self.num_files_completed = len(self.files_completed)
        total_bytes_read = self.total_bytes_read_completed_files
        if self.current_file_bytes_read:
            total_bytes_read += self.current_file_bytes_read
        elif (
            self.current_file_elements_read
            and self.current_file_num_elements
            and self.current_file_size
        ):
            current_files_bytes_read = int(
                (self.current_file_elements_read / self.current_file_num_elements)
                * self.current_file_size
            )
            total_bytes_read += current_files_bytes_read
        percentage_completion = (
            100.0 * (total_bytes_read / self.total_bytes_on_disk)
            if self.total_bytes_on_disk
            else -1
        )

        if percentage_completion > 0:
            self.estimated_time_to_completion_in_minutes = int(
                (
                    self.running_time.total_seconds()
                    * (100 - percentage_completion)
                    / percentage_completion
                )
                / 60
            )
            self.percentage_completion = f"{percentage_completion:.2f}%"


@platform_name("Metadata File")
@config_class(FileSourceConfig)
@support_status(SupportStatus.CERTIFIED)
class GenericFileSource(StatefulIngestionSourceBase, TestableSource):
    """
    This plugin pulls metadata from a previously generated file.
    The [metadata file sink](../../../../metadata-ingestion/sink_docs/metadata-file.md) can produce such files, and a number of
    samples are included in the [examples/mce_files](../../../../metadata-ingestion/examples/mce_files) directory.
    """

    def __init__(self, ctx: PipelineContext, config: FileSourceConfig):
        super().__init__(config, ctx)
        self.ctx = ctx
        self.config = config
        self.report: FileSourceReport = FileSourceReport()

    @classmethod
    def create(cls, config_dict, ctx):
        config = FileSourceConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_filenames(self) -> Iterable[FileInfo]:
        path_str = str(self.config.path)
        schema = get_path_schema(path_str)
        fs_class = fs_registry.get(schema)
        fs = fs_class.create()
        for file_info in fs.list(path_str):
            if file_info.is_file and file_info.path.endswith(
                self.config.file_extension
            ):
                yield file_info

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        # No super() call, as we don't want helpers that create / remove workunits
        return [
            partial(auto_workunit_reporter, self.report),
            auto_status_aspect if self.config.stateful_ingestion else None,
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_workunits_internal(
        self,
    ) -> Iterable[MetadataWorkUnit]:
        for f in self.get_filenames():
            for i, obj in self.iterate_generic_file(f):
                id = f"{f.path}:{i}"
                if isinstance(
                    obj, (MetadataChangeProposalWrapper, MetadataChangeProposal)
                ):
                    if (
                        self.config.aspect is not None
                        and obj.aspectName is not None
                        and obj.aspectName != self.config.aspect
                    ):
                        continue

                    if isinstance(obj, MetadataChangeProposalWrapper):
                        yield MetadataWorkUnit(id, mcp=obj)
                    else:
                        yield MetadataWorkUnit(id, mcp_raw=obj)
                else:
                    yield MetadataWorkUnit(id, mce=obj)
            self.report.total_num_files += 1
            self.report.append_total_bytes_on_disk(f.size)

    def get_report(self):
        return self.report

    def close(self):
        super().close()

    def _iterate_file(self, file_status: FileInfo) -> Iterable[Any]:
        file_read_mode = self.config.read_mode
        if file_read_mode == FileReadMode.AUTO:
            if file_status.size < self.config._minsize_for_streaming_mode_in_bytes:
                file_read_mode = FileReadMode.BATCH
            else:
                file_read_mode = FileReadMode.STREAM

        # Open the file.
        schema = get_path_schema(file_status.path)
        fs_class = fs_registry.get(schema)
        fs = fs_class.create()
        self.report.current_file_name = file_status.path
        self.report.current_file_size = file_status.size
        fp = fs.open(file_status.path)

        with fp:
            if file_read_mode == FileReadMode.STREAM:
                yield from self._iterate_file_streaming(fp)
            else:
                yield from self._iterate_file_batch(fp)

        self.report.files_completed.append(file_status.path)
        self.report.num_files_completed += 1
        self.report.total_bytes_read_completed_files += self.report.current_file_size
        self.report.reset_current_file_stats()

    def _iterate_file_streaming(self, fp: Any) -> Iterable[Any]:
        # Count the number of elements in the file.
        if self.config.count_all_before_starting:
            count_start_time = datetime.datetime.now()
            parse_stream = ijson.parse(fp, use_float=True)
            total_elements = 0
            for _row in ijson.items(parse_stream, "item", use_float=True):
                total_elements += 1
            count_end_time = datetime.datetime.now()
            self.report.add_count_time(count_end_time - count_start_time)
            self.report.current_file_num_elements = total_elements
            fp.seek(0)

        # Read the file.
        self.report.current_file_elements_read = 0
        parse_start_time = datetime.datetime.now()
        parse_stream = ijson.parse(fp, use_float=True)
        for row in ijson.items(parse_stream, "item", use_float=True):
            parse_end_time = datetime.datetime.now()
            self.report.add_parse_time(parse_end_time - parse_start_time)
            self.report.current_file_elements_read += 1
            yield row

    def _iterate_file_batch(self, fp: Any) -> Iterable[Any]:
        # Read the file.
        contents = json.load(fp)

        # Maintain backwards compatibility with the single-object format.
        if isinstance(contents, list):
            for row in contents:
                yield row
        else:
            yield contents

    def iterate_mce_file(self, path: str) -> Iterator[MetadataChangeEvent]:
        # TODO: Remove this method, as it appears to be unused.
        schema = get_path_schema(path)
        fs_class = fs_registry.get(schema)
        fs = fs_class.create()
        file_status = fs.file_status(path)
        for obj in self._iterate_file(file_status):
            mce: MetadataChangeEvent = MetadataChangeEvent.from_obj(obj)
            yield mce

    def iterate_generic_file(
        self, file_status: FileInfo
    ) -> Iterator[
        Tuple[
            int,
            Union[
                MetadataChangeEvent,
                MetadataChangeProposalWrapper,
                MetadataChangeProposal,
            ],
        ]
    ]:
        for i, obj in enumerate(self._iterate_file(file_status)):
            try:
                deserialize_start_time = datetime.datetime.now()
                item = _from_obj_for_file(obj)
                if item is not None:
                    deserialize_duration = (
                        datetime.datetime.now() - deserialize_start_time
                    )
                    self.report.add_deserialize_time(deserialize_duration)
                    yield i, item
            except Exception as e:
                self.report.report_failure(f"path-{i}", str(e))

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        config = FileSourceConfig.parse_obj(config_dict)
        exists = os.path.exists(config.path)
        if not exists:
            return TestConnectionReport(
                basic_connectivity=CapabilityReport(
                    capable=False,
                    failure_reason=f"{config.path} doesn't appear to be a valid file or directory.",
                )
            )
        is_dir = os.path.isdir(config.path)
        failure_message = None
        readable = os.access(config.path, os.R_OK)
        if not readable:
            failure_message = f"Cannot read {config.path}"
        if is_dir:
            executable = os.access(config.path, os.X_OK)
            if not executable:
                failure_message = f"Do not have execute permissions in {config.path}"

        if failure_message:
            return TestConnectionReport(
                basic_connectivity=CapabilityReport(
                    capable=False, failure_reason=failure_message
                )
            )
        else:
            return TestConnectionReport(
                basic_connectivity=CapabilityReport(capable=True)
            )

    @staticmethod
    def close_if_possible(stream):
        if hasattr(stream, "close") and callable(stream.close):
            stream.close()


def _from_obj_for_file(
    obj: dict,
) -> Union[
    MetadataChangeEvent,
    MetadataChangeProposal,
    MetadataChangeProposalWrapper,
    None,
]:
    item: Union[
        MetadataChangeEvent,
        MetadataChangeProposalWrapper,
        MetadataChangeProposal,
        UsageAggregationClass,
    ]

    if "proposedSnapshot" in obj:
        item = MetadataChangeEvent.from_obj(obj)
    elif "aspect" in obj:
        item = MetadataChangeProposalWrapper.from_obj(obj)
    else:
        item = UsageAggregationClass.from_obj(obj)
    if not item.validate():
        raise ValueError(f"failed to parse: {obj}")

    if isinstance(item, UsageAggregationClass):
        logger.warning(f"Dropping deprecated UsageAggregationClass: {item}")
        return None
    else:
        return item


def read_metadata_file(
    file: pathlib.Path,
) -> Iterable[
    Union[
        MetadataChangeEvent,
        MetadataChangeProposal,
        MetadataChangeProposalWrapper,
    ]
]:
    # This simplified version of the FileSource can be used for testing purposes.
    with file.open("r") as f:
        for obj in json.load(f):
            item = _from_obj_for_file(obj)
            if item:
                yield item



import logging
from functools import partial
from typing import Any, Dict, List, Optional, Union, cast

from lark import Token, Tree

from datahub.ingestion.source.powerbi.m_query.data_classes import (
    TRACE_POWERBI_MQUERY_PARSER,
)

logger = logging.getLogger(__name__)


def get_output_variable(root: Tree) -> Optional[str]:
    in_expression_tree: Optional[Tree] = get_first_rule(root, "in_expression")
    if in_expression_tree is None:
        return None
    # Get list of terminal value
    # Remove any whitespaces
    # Remove any spaces
    return "".join(
        strip_char_from_list(
            remove_whitespaces_from_list(token_values(in_expression_tree)), " "
        )
    )


def get_variable_statement(parse_tree: Tree, variable: str) -> Optional[Tree]:
    _filter = parse_tree.find_data("variable")
    # filter will return statement of the form <variable-name> = <expression>
    # We are searching for Tree where variable-name is matching with provided variable
    for tree in _filter:
        values: List[str] = token_values(tree.children[0])
        actual_value: str = "".join(strip_char_from_list(values, " "))
        if TRACE_POWERBI_MQUERY_PARSER:
            logger.debug(f"Actual Value = {actual_value}")
            logger.debug(f"Expected Value = {variable}")

        if actual_value.lower() == variable.lower():
            return tree

    logger.debug(f"Provided variable({variable}) not found in variable rule")

    return None


def get_first_rule(tree: Tree, rule: str) -> Optional[Tree]:
    """
    Lark library doesn't have an advance search function.
    This function will return the first tree of the provided rule
    :param tree: Tree to search for the expression rule
    :return: Tree
    """

    def internal(node: Union[Tree, Token]) -> Optional[Tree]:
        if isinstance(node, Tree) and node.data == rule:
            return node
        if isinstance(node, Token):
            return None

        for child in cast(Tree, node).children:
            child_node: Optional[Tree] = internal(child)
            if child_node is not None:
                return child_node

        return None

    expression_tree: Optional[Tree] = internal(tree)

    return expression_tree


def token_values(tree: Tree, parameters: Dict[str, str] = {}) -> List[str]:
    """
    :param tree: Tree to traverse
    :param parameters: If parameters is not an empty dict, it will try to resolve identifier variable references
                       using the values in 'parameters'.
    :return: List of leaf token data
    """
    values: List[str] = []

    def internal(node: Union[Tree, Token]) -> None:
        if parameters and isinstance(node, Tree) and node.data == "identifier":
            # This is the case where they reference a variable using
            # the `#"Name of variable"` or `Variable` syntax. It can be
            # a quoted_identifier or a regular_identifier.

            ref = make_function_name(node)

            # For quoted_identifier, ref will have quotes around it.
            if ref.startswith('"') and ref[1:-1] in parameters:
                resolved = parameters[ref[1:-1]]
                values.append(resolved)
            elif ref in parameters:
                resolved = parameters[ref]
                values.append(resolved)
            else:
                # If we can't resolve, fall back to the name of the variable.
                logger.debug(f"Unable to resolve parameter reference to {ref}")
                values.append(ref)
        elif isinstance(node, Token):
            values.append(cast(Token, node).value)
            return
        else:
            for child in node.children:
                internal(child)

    internal(tree)

    return values


def remove_whitespaces_from_list(values: List[str]) -> List[str]:
    result: List[str] = []
    for item in values:
        if item.strip() not in ("", "\n", "\t"):
            result.append(item)

    return result


def strip_char(value: str, char: str = '"') -> str:
    return value.strip(char)


def strip_char_from_list(values: List[str], char: str = '"') -> List[str]:
    result: List[str] = []
    for item in values:
        result.append(strip_char(item.strip(char), char=char))

    return result


def make_function_name(tree: Tree) -> str:
    values: List[str] = token_values(tree)
    return ".".join(values)


def get_all_function_name(tree: Tree) -> List[str]:
    """
    Returns all function name present in an input tree
    :param tree: Input lexical tree
    :return: list of function name
    """
    functions: List[str] = []

    # List the all invoke_expression in the Tree
    _filter: Any = tree.find_data("invoke_expression")

    for node in _filter:
        if TRACE_POWERBI_MQUERY_PARSER:
            logger.debug(f"Tree = {node.pretty()}")
        primary_expression_node: Optional[Tree] = first_primary_expression_func(node)
        if primary_expression_node is None:
            continue

        identifier_node: Optional[Tree] = first_identifier_func(primary_expression_node)
        if identifier_node is None:
            continue

        functions.append(make_function_name(identifier_node))

    return functions


def flat_argument_list(tree: Tree) -> List[Tree]:
    values: List[Tree] = []

    for child in tree.children:
        if isinstance(child, Token):
            continue
        if isinstance(child, Tree) and (
            child.data == "argument_list" or child.data == "expression"
        ):
            values.append(child)

    return values


first_expression_func = partial(get_first_rule, rule="expression")
first_item_selector_func = partial(get_first_rule, rule="item_selector")
first_arg_list_func = partial(get_first_rule, rule="argument_list")
first_identifier_func = partial(get_first_rule, rule="identifier")
first_primary_expression_func = partial(get_first_rule, rule="primary_expression")
first_invoke_expression_func = partial(get_first_rule, rule="invoke_expression")
first_type_expression_func = partial(get_first_rule, rule="type_expression")
first_list_expression_func = partial(get_first_rule, rule="list_expression")



import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Iterable, List, Optional, Tuple, Type, Union, ValuesView

import bson.timestamp
import pymongo.collection
from packaging import version
from pydantic import PositiveInt, validator
from pydantic.fields import Field
from pymongo.mongo_client import MongoClient

from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.emitter.mce_builder import (
    make_data_platform_urn,
    make_dataplatform_instance_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import (
    DatabaseKey,
    add_dataset_to_container,
    gen_containers,
)
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import DatasetContainerSubTypes
from datahub.ingestion.source.schema_inference.object import (
    SchemaDescription,
    construct_schema,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
    StaleEntityRemovalSourceReport,
    StatefulIngestionConfigBase,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.metadata.schema_classes import (
    ArrayTypeClass,
    BooleanTypeClass,
    BytesTypeClass,
    DataPlatformInstanceClass,
    DatasetPropertiesClass,
    NullTypeClass,
    NumberTypeClass,
    RecordTypeClass,
    SchemaFieldClass as SchemaField,
    SchemaFieldDataTypeClass as SchemaFieldDataType,
    SchemalessClass,
    SchemaMetadataClass as SchemaMetadata,
    StringTypeClass,
    TimeTypeClass,
    UnionTypeClass,
)
from datahub.metadata.urns import DatasetUrn

logger = logging.getLogger(__name__)

# These are MongoDB-internal databases, which we want to skip.
# See https://docs.mongodb.com/manual/reference/local-database/ and
# https://docs.mongodb.com/manual/reference/config-database/ and
# https://stackoverflow.com/a/48273736/5004662.
DENY_DATABASE_LIST = {"admin", "config", "local"}


class HostingEnvironment(Enum):
    SELF_HOSTED = "SELF_HOSTED"
    ATLAS = "ATLAS"
    AWS_DOCUMENTDB = "AWS_DOCUMENTDB"


class MongoDBConfig(
    PlatformInstanceConfigMixin, EnvConfigMixin, StatefulIngestionConfigBase
):
    # See the MongoDB authentication docs for details and examples.
    # https://pymongo.readthedocs.io/en/stable/examples/authentication.html
    connect_uri: str = Field(
        default="mongodb://localhost", description="MongoDB connection URI."
    )
    username: Optional[str] = Field(default=None, description="MongoDB username.")
    password: Optional[str] = Field(default=None, description="MongoDB password.")
    authMechanism: Optional[str] = Field(
        default=None, description="MongoDB authentication mechanism."
    )
    options: dict = Field(
        default={}, description="Additional options to pass to `pymongo.MongoClient()`."
    )
    enableSchemaInference: bool = Field(
        default=True, description="Whether to infer schemas. "
    )
    schemaSamplingSize: Optional[PositiveInt] = Field(
        default=1000,
        description="Number of documents to use when inferring schema size. If set to `null`, all documents will be scanned.",
    )
    useRandomSampling: bool = Field(
        default=True,
        description="If documents for schema inference should be randomly selected. If `False`, documents will be selected from start.",
    )
    maxSchemaSize: Optional[PositiveInt] = Field(
        default=300, description="Maximum number of fields to include in the schema."
    )
    # mongodb only supports 16MB as max size for documents. However, if we try to retrieve a larger document it
    # errors out with "16793600" as the maximum size supported.
    maxDocumentSize: Optional[PositiveInt] = Field(default=16793600, description="")

    hostingEnvironment: Optional[HostingEnvironment] = Field(
        default=HostingEnvironment.SELF_HOSTED,
        description="Hosting environment of MongoDB, default is SELF_HOSTED, currently support `SELF_HOSTED`, `ATLAS`, `AWS_DOCUMENTDB`",
    )

    database_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="regex patterns for databases to filter in ingestion.",
    )
    collection_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="regex patterns for collections to filter in ingestion.",
    )
    # Custom Stateful Ingestion settings
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = None

    @validator("maxDocumentSize")
    def check_max_doc_size_filter_is_valid(cls, doc_size_filter_value):
        if doc_size_filter_value > 16793600:
            raise ValueError("maxDocumentSize must be a positive value <= 16793600.")
        return doc_size_filter_value


@dataclass
class MongoDBSourceReport(StaleEntityRemovalSourceReport):
    filtered: List[str] = field(default_factory=list)

    def report_dropped(self, name: str) -> None:
        self.filtered.append(name)


# map PyMongo types to canonical MongoDB strings
PYMONGO_TYPE_TO_MONGO_TYPE = {
    list: "ARRAY",
    dict: "OBJECT",
    type(None): "null",
    bool: "boolean",
    int: "integer",
    bson.int64.Int64: "biginteger",
    float: "float",
    str: "string",
    bson.datetime.datetime: "date",
    bson.timestamp.Timestamp: "timestamp",
    bson.dbref.DBRef: "dbref",
    bson.objectid.ObjectId: "oid",
    bson.Decimal128: "numberDecimal",
    "mixed": "mixed",
}

# map PyMongo types to DataHub classes
_field_type_mapping: Dict[Union[Type, str], Type] = {
    list: ArrayTypeClass,
    bool: BooleanTypeClass,
    type(None): NullTypeClass,
    int: NumberTypeClass,
    bson.int64.Int64: NumberTypeClass,
    float: NumberTypeClass,
    str: StringTypeClass,
    bson.datetime.datetime: TimeTypeClass,
    bson.timestamp.Timestamp: TimeTypeClass,
    bson.dbref.DBRef: BytesTypeClass,
    bson.objectid.ObjectId: BytesTypeClass,
    bson.Decimal128: NumberTypeClass,
    dict: RecordTypeClass,
    "mixed": UnionTypeClass,
}


def construct_schema_pymongo(
    collection: pymongo.collection.Collection,
    delimiter: str,
    use_random_sampling: bool,
    max_document_size: int,
    should_add_document_size_filter: bool,
    sample_size: Optional[int] = None,
) -> Dict[Tuple[str, ...], SchemaDescription]:
    """
    Calls construct_schema on a PyMongo collection.

    Returned schema is keyed by tuples of nested field names, with each
    value containing 'types', 'count', 'nullable', 'delimited_name', and 'type' attributes.

    Parameters
    ----------
        collection:
            the PyMongo collection
        delimiter:
            string to concatenate field names by
        use_random_sampling:
            boolean to indicate if random sampling should be added to aggregation
        max_document_size:
            maximum size of the document that will be considered for generating the schema.
        should_add_document_size_filter:
            boolean to indicate if document size filter should be added to aggregation
        sample_size:
            number of items in the collection to sample
            (reads entire collection if not provided)
    """

    aggregations: List[Dict] = []
    if should_add_document_size_filter:
        doc_size_field = "temporary_doc_size_field"
        # create a temporary field to store the size of the document. filter on it and then remove it.
        aggregations = [
            {"$addFields": {doc_size_field: {"$bsonSize": "$$ROOT"}}},
            {"$match": {doc_size_field: {"$lt": max_document_size}}},
            {"$project": {doc_size_field: 0}},
        ]
    if use_random_sampling:
        # get sample documents in collection
        if sample_size:
            aggregations.append({"$sample": {"size": sample_size}})
        documents = collection.aggregate(
            aggregations,
            allowDiskUse=True,
        )
    else:
        if sample_size:
            aggregations.append({"$limit": sample_size})
        documents = collection.aggregate(aggregations, allowDiskUse=True)

    return construct_schema(list(documents), delimiter)


@platform_name("MongoDB")
@config_class(MongoDBConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.PLATFORM_INSTANCE, "Enabled by default")
@capability(SourceCapability.SCHEMA_METADATA, "Enabled by default")
@dataclass
class MongoDBSource(StatefulIngestionSourceBase):
    """
    This plugin extracts the following:

    - Databases and associated metadata
    - Collections in each database and schemas for each collection (via schema inference)

    By default, schema inference samples 1,000 documents from each collection. Setting `schemaSamplingSize: null` will scan the entire collection.
    Moreover, setting `useRandomSampling: False` will sample the first documents found without random selection, which may be faster for large collections.

    Note that `schemaSamplingSize` has no effect if `enableSchemaInference: False` is set.

    Really large schemas will be further truncated to a maximum of 300 schema fields. This is configurable using the `maxSchemaSize` parameter.

    """

    config: MongoDBConfig
    report: MongoDBSourceReport
    mongo_client: MongoClient
    platform: str = "mongodb"

    def __init__(self, ctx: PipelineContext, config: MongoDBConfig):
        super().__init__(config, ctx)
        self.config = config
        self.report = MongoDBSourceReport()

        options = {}
        if self.config.username is not None:
            options["username"] = self.config.username
        if self.config.password is not None:
            options["password"] = self.config.password
        if self.config.authMechanism is not None:
            options["authMechanism"] = self.config.authMechanism
        options = {
            **options,
            **self.config.options,
        }

        # See https://pymongo.readthedocs.io/en/stable/examples/datetimes.html#handling-out-of-range-datetimes
        self.mongo_client = MongoClient(
            self.config.connect_uri, datetime_conversion="DATETIME_AUTO", **options  # type: ignore
        )

        # This cheaply tests the connection. For details, see
        # https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html#pymongo.mongo_client.MongoClient
        self.mongo_client.admin.command("ping")

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "MongoDBSource":
        config = MongoDBConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_pymongo_type_string(
        self, field_type: Union[Type, str], collection_name: str
    ) -> str:
        """
        Return Mongo type string from a Python type

        Parameters
        ----------
            field_type:
                type of a Python object
            collection_name:
                name of collection (for logging)
        """
        try:
            type_string = PYMONGO_TYPE_TO_MONGO_TYPE[field_type]
        except KeyError:
            self.report.warning(
                message="Unrecognized column types found",
                context=f"Collection: {collection_name}, field type {field_type}",
            )
            PYMONGO_TYPE_TO_MONGO_TYPE[field_type] = "unknown"
            type_string = "unknown"

        return type_string

    def get_field_type(
        self, field_type: Union[Type, str], collection_name: str
    ) -> SchemaFieldDataType:
        """
        Maps types encountered in PyMongo to corresponding schema types.

        Parameters
        ----------
            field_type:
                type of a Python object
            collection_name:
                name of collection (for logging)
        """
        TypeClass: Optional[Type] = _field_type_mapping.get(field_type)

        if TypeClass is None:
            self.report.warning(
                message="Unrecognized column type found",
                context=f"Collection: {collection_name}, field type {field_type}",
            )
            TypeClass = NullTypeClass

        return SchemaFieldDataType(type=TypeClass())

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        database_names: List[str] = self.mongo_client.list_database_names()

        # traverse databases in sorted order so output is consistent
        for database_name in sorted(database_names):
            if database_name in DENY_DATABASE_LIST:
                continue
            if not self.config.database_pattern.allowed(database_name):
                self.report.report_dropped(database_name)
                continue

            database = self.mongo_client[database_name]
            database_key = DatabaseKey(
                database=database_name,
                platform=self.platform,
                instance=self.config.platform_instance,
                env=self.config.env,
            )
            yield from gen_containers(
                container_key=database_key,
                name=database_name,
                sub_types=[DatasetContainerSubTypes.DATABASE],
            )

            collection_names: List[str] = database.list_collection_names()
            # traverse collections in sorted order so output is consistent
            for collection_name in sorted(collection_names):
                dataset_name = f"{database_name}.{collection_name}"

                if not self.config.collection_pattern.allowed(dataset_name):
                    self.report.report_dropped(dataset_name)
                    continue

                dataset_urn = DatasetUrn.create_from_ids(
                    platform_id=self.platform,
                    table_name=dataset_name,
                    env=self.config.env,
                    platform_instance=self.config.platform_instance,
                )

                # Initialize data_platform_instance with a default value
                data_platform_instance = None
                if self.config.platform_instance:
                    data_platform_instance = DataPlatformInstanceClass(
                        platform=make_data_platform_urn(self.platform),
                        instance=make_dataplatform_instance_urn(
                            self.platform, self.config.platform_instance
                        ),
                    )

                dataset_properties = DatasetPropertiesClass(
                    name=collection_name,
                    tags=[],
                    customProperties={},
                )

                schema_metadata: Optional[SchemaMetadata] = None
                if self.config.enableSchemaInference:
                    schema_metadata = self._infer_schema_metadata(
                        collection=database[collection_name],
                        dataset_urn=dataset_urn,
                        dataset_properties=dataset_properties,
                    )

                # TODO: use list_indexes() or index_information() to get index information
                # See https://pymongo.readthedocs.io/en/stable/api/pymongo/collection.html#pymongo.collection.Collection.list_indexes.

                yield from add_dataset_to_container(database_key, dataset_urn.urn())
                yield from [
                    mcp.as_workunit()
                    for mcp in MetadataChangeProposalWrapper.construct_many(
                        entityUrn=dataset_urn.urn(),
                        aspects=[
                            schema_metadata,
                            dataset_properties,
                            data_platform_instance,
                        ],
                    )
                ]

    def _infer_schema_metadata(
        self,
        collection: pymongo.collection.Collection,
        dataset_urn: DatasetUrn,
        dataset_properties: DatasetPropertiesClass,
    ) -> SchemaMetadata:
        assert self.config.maxDocumentSize is not None
        collection_schema = construct_schema_pymongo(
            collection,
            delimiter=".",
            use_random_sampling=self.config.useRandomSampling,
            max_document_size=self.config.maxDocumentSize,
            should_add_document_size_filter=self.should_add_document_size_filter(),
            sample_size=self.config.schemaSamplingSize,
        )

        # initialize the schema for the collection
        canonical_schema: List[SchemaField] = []
        max_schema_size = self.config.maxSchemaSize
        collection_schema_size = len(collection_schema.values())
        collection_fields: Union[
            List[SchemaDescription], ValuesView[SchemaDescription]
        ] = collection_schema.values()
        assert max_schema_size is not None
        if collection_schema_size > max_schema_size:
            # downsample the schema, using frequency as the sort key
            self.report.report_warning(
                title="Too many schema fields",
                message=f"Downsampling the collection schema because it has too many schema fields. Configured threshold is {max_schema_size}",
                context=f"Schema Size: {collection_schema_size}, Collection: {dataset_urn}",
            )
            # Add this information to the custom properties so user can know they are looking at downsampled schema
            dataset_properties.customProperties["schema.downsampled"] = "True"
            dataset_properties.customProperties[
                "schema.totalFields"
            ] = f"{collection_schema_size}"

        logger.debug(f"Size of collection fields = {len(collection_fields)}")
        # append each schema field (sort so output is consistent)
        for schema_field in sorted(
            collection_fields,
            key=lambda x: (
                -x["count"],
                x["delimited_name"],
            ),  # Negate `count` for descending order, `delimited_name` stays the same for ascending
        )[0:max_schema_size]:
            field = SchemaField(
                fieldPath=schema_field["delimited_name"],
                nativeDataType=self.get_pymongo_type_string(
                    schema_field["type"], dataset_urn.name
                ),
                type=self.get_field_type(schema_field["type"], dataset_urn.name),
                description=None,
                nullable=schema_field["nullable"],
                recursive=False,
            )
            canonical_schema.append(field)

        # create schema metadata object for collection
        return SchemaMetadata(
            schemaName=collection.name,
            platform=f"urn:li:dataPlatform:{self.platform}",
            version=0,
            hash="",
            platformSchema=SchemalessClass(),
            fields=canonical_schema,
        )

    def is_server_version_gte_4_4(self) -> bool:
        try:
            server_version = self.mongo_client.server_info().get("versionArray")
            if server_version:
                logger.info(
                    f"Mongodb version for current connection - {server_version}"
                )
                server_version_str_list = [str(i) for i in server_version]
                required_version = "4.4"
                return version.parse(
                    ".".join(server_version_str_list)
                ) >= version.parse(required_version)
        except Exception as e:
            logger.error("Error while getting version of the mongodb server %s", e)

        return False

    def is_hosted_on_aws_documentdb(self) -> bool:
        return self.config.hostingEnvironment == HostingEnvironment.AWS_DOCUMENTDB

    def should_add_document_size_filter(self) -> bool:
        # the operation $bsonsize is only available in server version greater than 4.4
        # and is not supported by AWS DocumentDB, we should only add this operation to
        # aggregation for mongodb that doesn't run on AWS DocumentDB and version is greater than 4.4
        # https://docs.aws.amazon.com/documentdb/latest/developerguide/mongo-apis.html
        return (
            self.is_server_version_gte_4_4() and not self.is_hosted_on_aws_documentdb()
        )

    def get_report(self) -> MongoDBSourceReport:
        return self.report

    def close(self):
        self.mongo_client.close()
        super().close()



from typing import Generic, List, TypeVar, Union

from datahub.emitter.mcp_patch_builder import MetadataPatchProposal
from datahub.metadata.schema_classes import StructuredPropertyValueAssignmentClass
from datahub.utilities.urns.structured_properties_urn import (
    make_structured_property_urn,
)

_Parent = TypeVar("_Parent", bound=MetadataPatchProposal)


class StructuredPropertiesPatchHelper(Generic[_Parent]):
    def __init__(
        self,
        parent: _Parent,
        aspect_name: str = "structuredProperties",
    ) -> None:
        self.aspect_name = aspect_name
        self._parent = parent
        self.aspect_field = "properties"

    def parent(self) -> _Parent:
        return self._parent

    def set_property(
        self, key: str, value: Union[str, float, List[Union[str, float]]]
    ) -> "StructuredPropertiesPatchHelper":
        self.remove_property(key)
        self.add_property(key, value)
        return self

    def remove_property(self, key: str) -> "StructuredPropertiesPatchHelper":
        self._parent._add_patch(
            self.aspect_name,
            "remove",
            path=(self.aspect_field, make_structured_property_urn(key)),
            value={},
        )
        return self

    def add_property(
        self, key: str, value: Union[str, float, List[Union[str, float]]]
    ) -> "StructuredPropertiesPatchHelper":
        self._parent._add_patch(
            self.aspect_name,
            "add",
            path=(self.aspect_field, make_structured_property_urn(key)),
            value=StructuredPropertyValueAssignmentClass(
                propertyUrn=make_structured_property_urn(key),
                values=value if isinstance(value, list) else [value],
            ),
        )
        return self
